"""
    Copyright 2008-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys, functools
from . import eggwrapper as _eggwrapper, random

class Node(object):
    """
    Manage a single node of a given tree. This class provides an 
    interface to :class:`~.Tree` instance's nodes and allows access and 
    modification of data attached to a given node as well as the tree 
    descending from that node. A node must be understood as the point 
    *below* a branch (in the direction towards leaves). So the length
    describe in the instance concerns the branch *above* the
    corresponding node (towards the base of the tree). Branches 
    (connections between nodes) have a direction: they go *from* a node 
    *to* another node. Nodes have therefore *children*  and *parents* 
    (a given node have one parent except the base of the tree which has 
    none). Connecting a node to itself, making a two-way branch (to 
    branches connecting the same two nodes in opposite directions) or 
    duplicate branches (between the same two nodes and in the same 
    direction) are illegal.

    :param label: node label (in case of a terminal node, its leaf
        label), if needed. Labels, if provided, are expected to be
        strings for terminal nodes, and numeric values for internal
        nodes, but technically, all user-supplied values are accepted
        (however, some :class:`~.Tree` methods require proper types).
    """

    def __init__(self, label=None):
        self._parent = None
        self._parent_branch = None
        self._children = []
        self._child_branches = {}
        self._label = label

    def newick(self, skip_labels=False, skip_brlens=False):
        """
        Formats the node and the subtree descending from is as a
        newick string.

        :param skip_labels: omit internal branch labels.
        :param skip_brlens: omit the branch lengths.
        """
        return ''.join(self._newick(skip_labels, skip_brlens) + [';'])

    def _newick(self, skip_labels, skip_brlens):
        string_bits = []
        if len(self._children):
            string_bits.append('(')
            for child in self._children:
                string_bits.extend(child._newick(skip_labels, skip_brlens))
                brlen = self._child_branches[child]
                if brlen != None and not skip_brlens: string_bits.append(':' + str(brlen))
                string_bits.append(',')
            string_bits[-1] = ')'
        if self._label != None and (not skip_labels or not len(self._children)): string_bits.append(str(self._label))
        return string_bits

    def leaves_down(self):
        """
        Recursively gets all leaf labels descending from that node. If
        this is a terminal node, returns its label in a one-item list.
        """
        leaves = []
        if len(self._children) > 0:
            for child in self._children: leaves += child.leaves_down()
            return leaves
        else:
            return [self._label]

    def leaves_up(self):
        """
        Recursively gets all leaf labels contained on the other side of
        the tree. In other words, get all leaves of the tree
        except those descending from this
        node). If this is the root node, returns an empty list.
        """
        leaves = []
        if self._parent == None: return []
        else:
            for brother in self._parent._children:
                if brother != self: leaves += brother.leaves_down()
            leaves += self._parent.leaves_up()
        return leaves

    def _unlink(self):
        # Clears all references to other :class:`~.Node` instances
        # (parent and children) from this node. The nodes in question
        # are not affected (they are not deleted, and references they
        # hold to the current node are not erased).
        self._parent = None
        self._parent_branch = None
        self._children = []
        self._child_branches = {}

    def _add_child(self, label, brlen):
        # Generates a new :class:`~.Node` instance descending from the
        # current instance (add a child to this instance).
        # :param label: label of the new node.
        # :param brlen: length of the branch connecting this node to the
        #     new node.
        # :return: The new node.
        child = Node(label)
        child._parent = self
        child._parent_branch = brlen
        self._children.append(child)
        self._child_branches[child] = brlen
        return child

    def _connect(self, node, brlen):
        # Connect this node to an other, existing, node. The orientation
        # of the link is *from* the current instance *to* the passed
        # instance. If the passed node has children, they will be
        # retained, allowing to connect a subtree to this tree. It is
        # not allowed to connect a node that has a (non-``None`) parent.
        # To do so, one must first disconnect the node from its parent
        # and then connect it to a new one.
        # :param node: an existing node.
        # :param brlen: length of the branch between this node and the
        #     passed one.
        # tests are not performed for a hidden method
        # if node._parent is not None: raise ValueError('cannot connect node): already has a parent (disconnect first)'
        self._children.append(node)
        self._child_branches[node] = brlen
        node._parent = self
        node._parent_branch = brlen

    def has_descendant(self, node):
        """
        Tell if a :class:`.Node` is a descendant of this node.
        """
        for child in self._children:
            if child == node: return True
            if child.has_descendant(node): return True
        return False

    def is_parent(self, node):
        """
        Tell if a :class:`.Node` is the parent of this node.
        Passing ``None`` to an instance that has no parent will return
        ``True``.
        """
        return self._parent == node
    
    def is_child(self, node):
        """
        Tell if a :class:`.Node` is a child of this node.
        """
        for child in self._children:
            if node == child: return True
        return False

    @property
    def parent(self):
        """
        Parent of this node. ``None`` if the
        node has no parent.
        """
        return self._parent

    def children(self):
        """
        Return an iterator over this node's children.
        """
        for node in self._children: yield node

    def child(self, idx):
        """
        Return a given child, as a :class:`~.Node` instance.
        """
        try: return self._children[idx]
        except IndexError: raise IndexError('node index out of range')

    def siblings(self):
        """
        List of other children of this node's parent. It is required
        that this node has a parent.
        """
        if self._parent == None: raise ValueError('cannot access siblings: this node has no parent')
        return [node for node in self._parent._children if node != self]

    def _remove_parent(self):
        # Removes the branch between this node and its parent. Note that
        # this method removes also this node from its parent's children
        # (both sides of the branch are removed).
        self._parent._children.remove(self)
        del self._parent._child_branches[self]
        self._parent = None
        self._parent_branch = None

    def _remove_child(self, node):
        # Removes the branch between this node and one of its children.
        # Note that this method also removes the parent of child in
        # question (both sides of the branch are removed).
        self._children.remove(node)
        del self._child_branches[node]
        node._parent = None
        node._parent_branch = None

    def set_branch_to(self, child, brlen):
        """
        Set the length of the branch to a child node.

        :param child: node whose branch should be resized. It can be
            represented by a direct reference (as a :class:`.Node`), or
            by its index in the children list. In that case, ensure that
            the index is currently valid.

        :param brlen: new branch length (it is allowed to pass ``None``).
        """
        if isinstance(child, Node):
            if child not in self._children: raise ValueError('invalid node: not part of the list of children')
            self._child_branches[child] = brlen
            child._parent_branch = brlen
        else:
            if child >= len(self._children): raise ValueError('invalid node index')
            self._child_branches[self._children[child]] = brlen
            self._children[child]._parent_branch = brlen

    def branch_to(self, child):
        """
        Length of the branch to a child node.
        Non-specified branch lengths are represented by ``None``.

        :param child: node whose branch length is requested. It can be
            represented by a direct reference (as a :class:`.Node`), or
            by its index in the children list. In that case, ensure that
            the index is currently valid.
        """
        if isinstance(child, Node):
            if child not in self._children: raise ValueError('invalid node: not part of the list of children')
            return self._child_branches[child]
        else:
            if child >= len(self._children): raise ValueError('invalid node index')
            return self._child_branches[self._children[child]]

    @property
    def parent_branch(self):
        """
        Length of the branch to the parent. Non-specified branch
        lengths are represented by ``None``. An exception is thrown
        upon accessing this attribute if
        this node has no parent. This attribute can be modified.
        """
        if self._parent is None: raise ValueError('this node has not parent')
        return self._parent_branch

    @parent_branch.setter
    def parent_branch(self, brlen):
        if self._parent == None: raise ValueError('this node has no parent')
        self._parent.set_branch_to(self, brlen) # will set both brlen at both sides

    @property
    def label(self):
        """ Node's label (modifiable). """
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    @property
    def num_children(self):
        """ Number of children connected to this node. """
        return len(self._children)

class Tree(object):
    """
    Editable genealogical or phylogenetic tree. A tree is a linked collection of nodes
    which all have one parent (except the ultimate base node) and any
    number of children.

    :param fname: name of a newick-formatted file containing the tree to import.
    :param string: newick-formatted string representing the the tree to import.
        If it not allowed to specified both *fname* and *string*.

    The instance can be initialized as an empty tree (with only a root 
    node), or from a newick-formatted string. By default, the string is 
    read from the file name passed as the *fname* argument to the 
    constructor, but it can be passed directly through the constructor 
    argument *string*. It is not allowed to set both *fname* and 
    *string* at the same time. If neither is specified, an empty tree, 
    containing only a bare base node, is created. The newick parser 
    expects a well-formed newick string (including the trailing 
    semicolon).

    Nodes are implemented as :class:`.Node` instances.  A node without 
    children is a *leaf* and otherwise it is *internal*. A node with 
    exactly one child is generally meaningless, but is allowed. All 
    nodes (internal nodes as well as leaves) have a *label* which in 
    the case of leaves can be used as a sample name. It is not possible 
    to apply a name *and* a label to leaf node, in agreement with the 
    newick format. All connections between nodes (*branches*) are 
    oriented and can have a length (although the lengths can be 
    omitted) but note that labels are applied to nodes, not branches. 
    All :class:`!Tree` instances have one base node which is the only 
    one allowed not to have a parent. Network-like structures are not 
    allowed (because nodes must have exactly one parent).

    Import and export to/from strings and files are in the 
    bracket-based `newick  <http://evolution.genetics.washington.edu/phylip/newicktree.html>`_ 
    format (the parser treats terminal node labels as strings, and 
    internal node labels as integers or, by default, floats). 
    :class:`!Tree` instances can be exported as a built-in :class:`str` 
    function by the syntax ``str(tree)`` or using the method 
    :meth:`~.Tree.newick`. Nodes also have a :meth:`~.Node.newick` 
    method.

    :class:`!Tree` instances are iterable. Three iterators are 
    provided: one is depth-first (:meth:`~.Tree.depth_iter`), another 
    is breath-first (:meth:`~.Tree.breadth_iter`) and one iterates on 
    terminal nodes (leaves) only (:meth:`~.Tree.iter_leaves`).
    """
    def  __init__(self, fname=None, string=None):
        self._base = Node()
        self._leaves = []
        self._nodes = [self._base]
        if fname != None and string != None:
            raise ValueError('cannot set both arguments of Tree constructor')
        if fname != None:
            f = open(fname)
            string = f.read()
            f.close()
        if string != None:
            try: string = string.translate(None, ' \n\r\t')
            except TypeError: string = string.translate(dict.fromkeys(map(ord, ' \n\r\t')))
            if len(string) == 0: raise ValueError('empty newick string')
            if string[0] != '(' or string[-2:]!= ');': raise ValueError('invalid newick string')
            self._parse(string[:-1], self._base)

    @staticmethod
    def _from_coalesce(coal, idx, fun):
        # Create a :class:`~.Tree` instance from the results of a
        # coalescent simulation.
        #
        # :param coal: a :cpp:class:`~.Coalesce` object which has run a
        #     valid simulation.
        # :param idx: the index of a tree (must be within range of
        #     :cpp:meth:`~.Coalesce.number_of_trees`).
        # :param fun: executable expression converting the label to a
        #     string.
        #
        # :return: A new :class:`~.Tree` instance.
        source = coal.tree(idx)
        root = source.root()
        tree = Tree.__new__(Tree)
        tree._base = Node()
        tree._leaves = []
        tree._nodes = [tree._base]
        tree._from_coalesce_helper(tree._base, source, root, fun)
        return tree

    def _from_coalesce_helper(self, node_dst, tree_src, node_src, fun):
        # Helper for _from_coalesce
        # :param node_dst: current node in self.
        # :param tree_src: the reference C++ tree.
        # :param node_src: current C++ node in reference tree
        # :param fun: like for :meth:`~._from_coalesce`.
        if node_src.is_terminal():
            self._leaves.append(node_dst)
            node_dst._label = fun(node_src.label())
        else:
            son1 = tree_src.node(node_src.son1())
            child1 = node_dst._add_child(None, son1.get_L())
            self._from_coalesce_helper(child1, tree_src, son1, fun)
            self._nodes.append(child1)
            son2 = tree_src.node(node_src.son2())
            child2 = node_dst._add_child(None, son2.get_L())
            self._from_coalesce_helper(child2, tree_src, son2, fun)
            self._nodes.append(child2)

    def copy(self, node=None):
        """
        Make a deep copy of the tree.
        Create a new instance of :class:`.Tree` that is a deep copy of
        a subtree of the the current tree, or a deep copy of it all.

        :param node: a :class:`.Node` instance (one of the nodes of
            the current tree) at the base of the subtree that should be
            copied. By default, or if ``None``, or the base of the tree
            is passed, the whole tree is copied. It is not allowed to
            pass a leaf.
        """
        if node == None or node == self._base: return self._copy_full()
        if node not in self._nodes: raise ValueError('cannot copy tree: node is not part of tree')
        if node in self._leaves: raise ValueError('cannot copy tree: node is terminal')
        clone = Tree.__new__(Tree)
        clone._base = Node()
        clone._nodes = [clone._base]
        clone._leaves = []
        self._copy_helper(clone, clone._base, node)
        return clone

    def _copy_helper(self, clone, dest_node, node):
        dest_node._label = node._label
        if node in self._leaves:
            clone._leaves.append(dest_node)
        for child in node._children:
            new_node = dest_node._add_child(child._label, child._parent_branch)
            self._copy_helper(clone, new_node, child)
            clone._nodes.append(new_node)

    def _copy_full(self):
        # create a new bare tree (all nodes are also disconnected)
        clone = Tree.__new__(Tree)
        clone._nodes = [Node() for i in range(len(self._nodes))]
        clone._leaves = []

        # locate the base by its index
        base_index = self._nodes.index(self._base)
        clone._base = clone._nodes[base_index]

        # connect nodes bases on their indexes
        for node, copy in zip(self._nodes, clone._nodes):
            copy._label = node._label                  # copy label
            copy._parent_branch = node._parent_branch    # copy parent brlen
            if node._parent == None:
                copy._parent = None
            else:
                idx = self._nodes.index(node._parent)
                copy._parent = clone._nodes[idx]       # copy parent
            for child in node._children:
                idx = self._nodes.index(child)
                child_copy = clone._nodes[idx]
                copy._children.append(child_copy)     # copy a child
                copy._child_branches[child_copy] = node._child_branches[child] # copy child brlen
            if node in self._leaves: clone._leaves.append(copy)

        return clone

    def extract(self, node, label=None):
        """
        Remove a subtree. The subtree is returned as a new tree.
        All nodes of the subtree descending
        from the requested node will now belong to the new tree. The
        label of the node at the base of the selected clade is deleted.
        In the original tree, the extracted clade is replaced by a
        terminal node which has, by default, the label of the node at
        the base of the extracted clade, or the label passed as
        argument.

        :param node: a :class:`.Node` instance (one of the nodes of
            the current tree) at the base of the subtree that should be
            extracted. It is not allowed to pass a leaf or the base of
            the tree.

        :param label: label to affect to the terminal node that is
            introduced to replace the extracted clade in the orginal
            tree. By default (or if ``None``), use the label of the node
            passed as first argument (note that this label should be in
            principle a number).

        :return: A new :class:`.Tree` instance.
        """
        if node not in self._nodes: raise ValueError('cannot extract subtree: node is not part of tree')
        if node in self._leaves: raise ValueError('cannot extract subtree: node is terminal')
        if node == self._base: raise ValueError('cannot extract subtree: node is base')

        parent = node._parent
        brlen = node._parent_branch
        node._remove_parent()
        if label == None: label = node._label
        parent._add_child(label, brlen)

        clone = Tree.__new__(Tree)
        clone._base = node
        clone._nodes = []
        clone._leaves = []
        self._extract_helper(clone, node)
        return clone

    def _extract_helper(self, clone, node):
        clone._nodes.append(node)
        self._nodes.remove(node)
        if node in self._leaves:
            clone._leaves.append(node)
            self._leaves.remove(node)
        else:
            for child in node._children:
                self._extract_helper(clone, child)

    def __del__(self):
        for node in self._nodes: node._unlink()
        
    def _parse(self, string, cur):
        # This parser expects a string starting with an open round
        # bracket and ending with a closed one.
        if not len(string): raise IOError('invalid newick string (empty string)')
        string= string[1:-1]
        i=0

        while(True):

            # case of a subtree
            subtree = []
            if string[i] == '(':

                # we gather everything until the matching closing bracket
                subtree.append(string[i])
                acc = 1
                while acc:
                    i+=1
                    if i == len(string): raise ValueError('invalid newick string')
                    subtree.append(string[i])
                    if string[i] == '(': acc+=1
                    if string[i] == ')': acc-=1
                i+=1

            # get all until the next comma or end of string
            buff = ''
            while i < len(string) and string[i] != ',':
                buff += string[i]
                i+=1

            # convert to label and brlen
            if not len(buff):
                label = None
                brlen = None
            else:
                tbuff = buff.split(':')
                if len(tbuff) == 1:
                    label = buff
                    brlen = None
                elif len(tbuff) == 2:
                    if len(tbuff[0]) == 0: label = None
                    else: label = tbuff[0]
                    try: brlen = float(tbuff[1])
                    except ValueError: raise ValueError('invalid newick string (invalid branch length: {0}'.format(tbuff[1]))
                else: raise ValueError('invalid newick string')

            # add the node (whichever internal or terminal)
            node = cur._add_child(label=label, brlen=brlen)
            self._nodes.append(node)

            # in case there is a subtree, it is recursively parsed
            if len(subtree):
                self._parse(''.join(subtree), node)
                # try to convert label to int of float (silent otherwise)
                if node._label != None:
                    try: node._label = int(node._label)
                    except ValueError:
                        try: node._label = float(node._label)
                        except ValueError: pass

            # else, cache the leaf
            else:
                self._leaves.append(node)

            # if we didn't reach the end of the string, we continue parsing
            if i == len(string): break
            else:
                i+=1
                continue

    def __str__(self):
        return self._base.newick()

    @property
    def num_nodes(self):
        """
        Total number of nodes in the tree. This number is never smaller
        than 1, even for empty trees.
        """
        return len(self._nodes)

    @property
    def num_leaves(self):
        """
        Number of terminal nodes in the tree. If the tree is empty (only
        the default base node), the number of leaves is 0.
        """
        return len(self._leaves)

    def newick(self, skip_labels=False, skip_brlens=False):
        """
        Return the newick-formatted string representing the instance.

        :param skip_labels: omit internal branch labels.
        :param skip_brlens: omit the branch lengths.
        """
        return self._base.newick(skip_labels, skip_brlens)

    @property
    def base(self):
        """
        Basal node of the tree. If the tree is unrooted, this is a
        trifurcation whose location should be considered as arbitrary,
        unless one the three clades below this node is the outgroup.
        If the tree is rooted, this is the root). This attribute is a
        :class:`.Node` instance which can be modified, but it cannot be
        replaced.
        """
        return self._base

    def add_node(self, parent, label=None, brlen=None):
        """
        Add a node to the tree.

        :param parent: one of the nodes of this instance, as a
            :class:`.Node` instance.

        :param label: node label which will be internal node label or
            leaf name according to the final structure of the tree. The
            new node has initially no children and is therefore a leaf
            until it is itself connected to a child (if ever).

        :param brlen: length of the branche connecting *parent* to the
            new node.

        :return: The new node as a :class:`.Node` reference.
        """
        if parent not in self._nodes: raise ValueError('cannot add node to a node that does not belong to this tree')
        child = parent._add_child(label, brlen)
        self._leaves.append(child)
        self._nodes.append(child)
        if parent in self._leaves:
            self._leaves.remove(parent)

    def iter_leaves(self):
        """
        Return an iterator over the leaves as :class:`.Node` instances.
        """
        for node in self._leaves: yield node

    def get_leaf(self, label):
        """
        Get a terminal node. Return the node (as a :class:`.Node` instance) that
        has the requested leaf label. If several nodes have this label,
        returns the first one. If no nodes have this label, returns
        ``None``.
        """
        for leaf in self._leaves:
            if leaf._label == label: return leaf
        else:
            return None

    def depth_iter(self, start=None):
        """
        Return a depth-first iterator. Iterate over the :class:`~.Node`
        instances of the trees, starting from the base but, then,
        following a depth-first order.

        :param start: start point of the iteration, as a :class:`.Node`
            instance of this tree. By default, start from the base of
            the tree.
        """
        if start is None: start = self._base
        elif start not in self._nodes: raise ValueError('node does not belong to this tree')
        return _tree_depth_first_iterator(self, start)

    def breadth_iter(self, start=None):
        """
        Return a breadth-first iterator. Iterate over the
        :class:`~.Node` instances of the trees, starting from the base
        but, then, following a breadth-first order.

        :param start: start point of the iteration, as a :class:`.Node`
            instance of this tree. By default, start from the base of
            the tree.
        """
        if start is None: start = self._base
        elif start not in self._nodes: raise ValueError('node does not belong to this tree')
        return _tree_breath_first_iterator(self, start)

    def total_length(self):
        """
        Compute the sum of branch lengths.
        All branch lengths must be defined (non-``None``),
        otherwise a :exc:`ValueError` will be raised.
        """
        L = 0
        for node in self._nodes:
            for child in node.children():
                l = node.branch_to(child)
                if l == None:
                    raise ValueError('cannot compute tree\'s length: at least one branch has no defined length')
                L += l
        return L

    def find_clade(self, names, ancestral=False, both_sides=False):
        """
        Check whether a group is one of the clades defined by the tree.

        The leaf names must be provided as an iterable (most logically, 
        a :class:`set`). Leaf names are normally :class:`str` 
        instances. All leaves must be present in the tree.

        If the*ancestral* is ``False``, search for the clade that 
        contains the provided list of names as descendant. There must 
        not be any other name amongst its descendants. If the tree is 
        unrooted and oriented in such a way that the a base lies within 
        the requested clade, it will not be detected. It is possible (if
        *both_sides* is ``True``) to allow searching for the 
        complement of the clade, thereby detecting the right clade even 
        if it is at the base of the tree. By default, if this situation 
        occurs, the clade will not be detected.

        If the option *ancestral* is ``True``, search for the most 
        recent common ancestor of all leaves specified in *names*. Use 
        of this method necessarily supposes that the tree is rooted 
        (however, there is no requirement regarding its shape such as 
        bifurcation at the base) and it is not allowed to *both_sides* 
        is ``True``. With this option, it is not possible to have a 
        ``None`` return value (since it is required that all leaves are 
        present in the tree, in the worse case the base of the tree is 
        returned).

        :param name: a :class:`set` (or compatible) specifying the
            requested leaves (as node labels, normally :class:`str`
            instances).

        :param ancestral: whether to look for the most recent common
            ancestral clade containing requested leaves (by default,
            look for the clade containing the exact same list of leaves).

        :param both_sides: only allowed if *ancestral* is ``False``.
            Look for both the requested list of leaves and its
            complement, allowing to detect a clade even if it is
            spanning the base of the tree.

        :return: The :class:`.Node` instance, if it exists, which has
            the exact same list of descendants than *taxa*. If no such
            clade is found, ``None``.

        .. warning::
            This method assumes that all leaf names of the tree are
            unique, as well as the list of names provided as argument.
            If this condition is not fulfilled, the right clade might
            not be found even if it exists.

        .. versionchanged:: 3.0.0
            Replaces previous methods :meth:`!findGroup`,
            :meth:`!findMonophyleticGroup`, :meth:`!smallest_group` and
            :meth:`!smallest_monophyleticGroup` with a modification of
            the underlying algorithm.
        """
        query = set(names)

        # capture errors or trivial cases
        if ancestral == True and both_sides == True: raise ValueError('cannot combine `ancestral` and `both_sides` options')
        if len(names) == 0: raise ValueError('cannot find clade: empty list of names')
        leaves_set = set([leaf._label for leaf in self._leaves])

        if query < leaves_set: pass
        elif query == leaves_set: return self._base # clade is the tree itself
        else: raise ValueError('cannot find clade: names are not all present in the tree')
        if both_sides: compl = leaves_set - query

        # find any leaf belonging to query
        for node in self._leaves:
            if node._label in query:
                query.discard(node._label)
                break

        # at least one of the names should be found
        else: raise RuntimeError('unexpected case in `Tree.find_clade()`: please report is as a bug')

        # climbs up the tree collecting other leaves from the query
        while node != None:

            # the descendant of the current node must have been all processed
            # they are all found to be part of the query (and removed from it)
            # if query completed, it is the right node
            if len(query) == 0: return node

            # there should be a parent because if node is the root, then we should have collected all queried leaves
            if node.parent == None: raise RuntimeError('unexpected case in `Tree.find_clade()`: please report is as a bug')

            # collect descendants of the parent (except those already processed through the current node)
            siblings = node.siblings()

            # move one level up
            node = node.parent

            for sib in siblings:
                for label in sib.leaves_down():

                    # discard leaves that are in query
                    if label in query:
                        query.discard(label)

                    # if non-query leaf found
                    else:

                        # if we are looking for the most recent common ancestor, it is fine
                        if ancestral == True: pass

                        # if the current node is the base, the clade might be spanning the base
                        # try the complement (if user allowed it)
                        elif node.parent == None and both_sides == True:
                            return self.find_clade(names=compl, both_sides=False)

                        # otherwise there is no such clade in this tree
                        else: return None

        # in principle, we never reach this point if the query
        raise RuntimeError('unexpected case in `Tree.find_clade()`: please report is as a bug')

    def collapse(self, node, ignore_len=False, ignore_label=False):
        # note: this API method is also used by remove_node() with explicit option values
        """
        Collapse a branch of the tree.

        :param node: :class:`~.Node` representing the branch to remove.
        :param ignore_len: don't try to transfer branch lengths to
            children.
        :param ignore_label: don't transfer label of the destroyed node
            to its parent.

        *node* represents the branch that must be removed from the tree
        (this node is destroyed in the process). It must be one of the
        nodes contained in the tree (as a :class:`~.Node` instance), but
        not the base of the tree. It cannot be an terminal node (leaf). 

        If *ignore_label* is not set to ``True``, the label of the
        destroyed node is transferred to the parent based on the
        following procedure: (1) if the destroyed node's label is
        ``None``, nothing is done; (2) if the destroyed node's parent's
        label is ``None``, the destroyed node's label is copied to its
        parent as is; (3) otherwise both labels are converted to strings
        (if they are not yet) and concatenated as in the string ``a;b``
        (where ``a`` is the parent's label and ``b`` is the destroyed
        node's label), even if the two labels are identical.

        If *ignore_len* is not set to ``True``, and if the length of the
        removed branch (branch from the specified node to its parent) is
        specified, it will be spread equally among the branches to its
        children (see example below). This requires that the branch
        length to all children are specified. If the removed branch has
        no specified length, nothing is done.

        Collapsing node ``[4]`` on the following tree::
        
             /------------------------------------------->[1]
             |
             |             /----------------------------->[3]
             |             |
             |----------->[2]             /-------------->[5]
             |             |              |
             |             \------------>[4]
            [0]                           |
             |                            \-------------->[6]
             |
             |              /---------------------------->[8]
             |              | 
             \------------>[7]            /------------->[10]
                            |             |
                            \----------->[9]
                                          |
                                          \------------->[11]
                                          
        will generate the following tree, with the correction of edge
        lengths as  depicted::

             /------------------------------------------->[1]
             |
             |             /----------------------------->[3]
             |             |
             |----------->[2]
             |             |
             |             |-------------------->[5]        L5 = L5+L4/2
            [0]            |
             |             \-------------------->[6]        L6 = L6+L4/2
             |
             |              /---------------------------->[8]
             |              | 
             \------------>[7]            /------------->[10]
                            |             |
                            \----------->[9]
                                          |
                                          \------------->[11]

        Although the total edge length of the tree is not modified, the
        relationships will be altered: the distance between the
        descendants of the collapsed node (nodes 5 and 6 in the example
        above) will be artificially increased.
        """

        # sanity checks
        if node not in self._nodes != self: raise ValueError('cannot collapse a node that does not belong to this tree')
        if node == self._base: raise ValueError('cannot collapse the base of the tree')
        if node.num_children == 0: raise ValueError('cannot collapse a terminal node')
        parent = node.parent
        children = list(node.children())

        # process branch lengths
        p_len = node.parent_branch
        s_len = [node.branch_to(child) for child in children]
        if not ignore_len and p_len != None:
            if None in s_len: raise ValueError('cannot collapse node: branches to children are required')
            s_len = [length + p_len/node.num_children for length in s_len]

        # disconnects the node to be collapsed and remove it
        node._remove_parent()
        for child in children: node._remove_child(child)
        self._nodes.remove(node)

        # connects the sons their grandparent
        for child, length in zip(children, s_len):
            parent._connect(child, length)

        # saves the label information
        if not ignore_label:
            plab = parent._label
            nlab = node._label
            if nlab != None:
                if plab == None: parent.label = nlab
                else: parent.label = '%s;%s' %(plab, nlab)

    def root(self, outgroup, branch_split=0.5, reoriente=False):
        """
        Root or reoriente the tree. By default, a new node is created
        to represent the root and is placed on the branch leading to
        the provided outgroup node (the second argument determines where
        the new node is placed on this branch). Otherwise, the tree is
        reoriented such as its base is placed at the location of the
        provided outgroup. In the former case, its ends with a
        bifurcation at the root; in the latter case, a trifurcation.
        It is illegal to call this method on trees that are already
        rooted (have a difurcation at the root).

        :param outgroup: :class:`~.Node` instance contained in this
            tree. It can be a leaf or any internal node, but not the
            current base of the tree (unless *reoriente* is ``True``: in
            that case, it *might* be the base of the tree [it will not
            change anything] and it *cannot* be a leaf).

        :param branch_split: where to cut the branch leading to the
            outgroup.

        :param reoriente: don't create any root node *branch_site* is
            therefore not considered) and only place the node provided
            as *outgroup* at the base of the tree, thereby
            merely changing the representation of the tree.

        The information below describes the case where
        *reoriente* is ``False`` (proper rooting).

        If the branch to the provided outgroup doesn't have a branch
        length, the *branch_split* argument is ignored. Otherwise,
        *branch_split* must be a real number between 0 and 1 and give
        the proportion of the branch that must be allocated to the basal
        branch leading to the outgroup, the complement being allocated
        to the branch leading to the rest of the tree. If *branch_split*
        is either 0 or 1, one of the branch will have a length of 0, but
        it will exist anyway.

        If the original tree has this structure::

             /------------------------------------------->[1]
             |
             |             /----------------------------->[3]
             |             |
             |----------->[2]             /-------------->[5]
             |             |              |
             |             \------------>[4]
            [0]                           |
             |                            \-------------->[6]
             |
             |              /---------------------------->[8]
             |              | 
             \------------>[7]            /------------->[10]
                            |             |
                            \---[ROOT]-->[9]
                                          |
                                          \------------->[11]

        And rooting is requested at node ``[9]``, the root will be
        placed on the branch marked by ``[ROOT]``.  The outcome will be
        as depicted below, with the introduction of a new node (marked
        ``[ROOT]``) and the reorientation of the tree to place it at the
        base::

                                /-------------------------[1]
                                |
                         /-----[0]     /------------------[3]
                         |      |      |
                         |      \-----[2]      /----------[5]
                         |             |       |
             /---[E2]---[7]            \------[4]
             |           |                     |
             |           |                     \----------[6]
             |           |
           [ROOT]        \--------------------------------[8]
             |
             |                     /---------------------[10]
             |                     |
             \--------[E1]--------[9]
                                   |
                                   \---------------------[11]

        In this example, the relationship between nodes ``[7]`` and
        ``[0]`` (the previous base of the tree) is reverted. The label
        of node ``[7]`` is automatically transferred to node ``[0]``.
        This is consistent with the idea that internal node labels
        describe a property of the branch. The original label of the
        base, if it exists, is discarded. Since the branch between
        ``[7]`` and ``[9]`` is cut in two, the original label of node
        ``[9]`` is copied to node ``[7]``, leaving them both with the
        same label. However, if the outgroup is a terminal node, the
        label is not copied and the other basal branch is left without
        label.

        Let :math:`L` be the length of the branch from ``[7]`` and ``[9]``
        in the original tree, and :math:`r` the value of the parameter
        *branch_split*. The length of the branch ``[E1]`` will be set
        to :math:`rL`, and the branch ``[E2]`` to :math:`(1-r)L`. Overall, the
        length of the tree will not be modified.

        In the case that *reoriente* is ``True``, the final tree is rather::

            /------------------------------------------->[10]
            |
            |------------------------------------------->[11]
            |
           [9]       /----------------------------------->[8]
            |        |
            \------>[7]       /-------------------------->[1]
                     |        |
                     \------>[0]      /------------------>[3]
                              |       |
                              \----->[2]        /-------->[5]
                                      |         |
                                      \------->[4]
                                                |
                                                \-------->[6]

        The topology of the tree is the same as the initial one, except
        that the base is now ``[9]``. The lengths of all branches are
        conserved. However, node labels between the old and the new base
        are reverted: the node label of the new base (``[9]`` in the
        example) is affected to the next node (``[7]`` in the example)
        and so on until the old base (``[0]`` in the example), whose
        label, if it exists, is discarded.
        """

        # sanity checking
        if outgroup not in self._nodes: raise ValueError('cannot root tree: outgroup node is not part of the tree')
        if outgroup == self._base and reoriente == False: raise ValueError('cannot root tree: outgroup node is the current base of the tree')
        if outgroup.num_children == 0 and reoriente == True: raise ValueError('cannot reoriente tree: outgroup node is a terminal node')
        if self._base.num_children == 2: raise ValueError('cannot root tree: tree is already rooted')
        if self._base.num_children < 2: raise ValueError('cannot root tree: tree has an invalid structure')
        L = outgroup.parent_branch
        if not reoriente and L != None and (branch_split < 0.0 or branch_split > 1.0): raise ValueError('cannot root tree: invalid `branch_split` value')

        # collect all nodes on the path from the new to the old root
        path = []
        labels = [] # collect also labels
        brlens = [] # and branch lengths

        if reoriente: cur = outgroup
        else: cur = outgroup.parent

        while cur != None:
            path.append(cur)
            labels.append(cur._label) # last one will be discarded
            brlens.append(cur._parent_branch) # last one is None
            cur = cur.parent
            # path[0] = node after new root (or new base if reoriente)
            # path[...] = intermediate nodes
            # path[-1] = base of the tree

        if not reoriente:

            # remove the branch where the root is placed
            outgroup._remove_parent()

            # create root and add it to nodes
            root = Node()
            self._nodes.append(root)
            self._base = root

        else:
            self._base = outgroup

        # disconnect all nodes on the way from new base to old base (old base not included)
        for node in path[:-1]: node._remove_parent()

        # reconnect in the other order
        for i in range(len(path)-1):
            path[i]._connect(path[i+1], brlens[i])  # note that brlens[-1] is discarded / is None anyway
            path[i+1]._label = labels[i]            # labels[-1] is discarded and path[0]._label unset yet

        if not reoriente:

            # compute length of basal branches
            if L != None:
                L1 = L * branch_split
                L2 = L * (1.0 - branch_split)
            else:
                L1 = None
                L2 = None

            # connect the new root to the two subtrees
            root._connect(outgroup, L1)
            root._connect(path[0], L2)

        if not reoriente:
            # copy the root branch label to the new branch
            if outgroup.num_children > 0: path[0]._label = outgroup._label
            else: path[0]._label = None
        else:
            outgroup._label = None

    def map_descendants(self):
        """
        Map all leaves of the trees to internal nodes.
        Generate and return a :class:`dict` which gives, for all internal
        nodes of the trees (excluding the base), the list of terminal
        nodes that are ultimately connected when reading the tree
        away from its base.
        """
        mapping = {}
        queue = {}

        # add the parent of all leaves to the queue
        for leaf in self._leaves:
            if leaf._parent not in queue: queue[leaf._parent] = []
            queue[leaf._parent].append([leaf._label])

        # process queue'd node and their parent until exhaustion
        while len(queue):
            to_del = set()
            to_add = {}

            for node in queue:
                items = queue[node]

                # this checks if the node has been fully processed
                if len(items) == len(node._children):

                    # if the node is the base of the tree, just trash the labels
                    if node._parent != None:

                        # otherwise, send the data to the mapping...
                        mapping[node] = functools.reduce(list.__add__, items)

                        # ... and propagate labels to higher level
                        if node._parent in queue: queue[node._parent].append(mapping[node])
                        elif node._parent in to_add: to_add[node._parent].append(mapping[node])
                        else: to_add[node._parent] = [mapping[node]]

                    # remove the node from the queue
                    to_del.add(node)

            for node in to_del: del queue[node]
            queue.update(to_add)

        for node in mapping: mapping[node] = tuple(mapping[node])
        return mapping

    def frequency_nodes(self, trees, relative=False):
        """
        Label nodes based on their number of occurrences in a list 
        of trees. Each node receives an integer as label counting the 
        number of trees where the same node exists among the trees 
        provided as argument. It is required that all leaf labels are 
        unique.

        :param trees: an iterable containing :class:`.Tree` instances
            with exactly the same set of leaf labels.
        :param relative: node frequencies are expressed as fractions.
            The use of this option requires that at least one tree is
            provided.

        .. note::
            With the exception of the base of the tree (which is ignored
            by this function) and leaf labels, all previously set labels
            are erased.
        """

        # prepare a set for safety checking
        leaves = [node._label for node in self._leaves]
        n = len(leaves)
        leaves = set(leaves)
        if len(leaves) != n: raise ValueError('cannot compute node frequencies: not all leaf labels are unique')

        # get own list of nodes and reverse relationships
        nodes = dict([(clade, node) for (node, clade) in self.map_descendants().items()])

        # initialize counter dict
        counts = dict.fromkeys([nodes[i] for i in nodes], 0)

        # process all provided trees
        n = 0
        for tree in trees:
            n += 1

            # check consistency
            if set([node._label for node in tree._leaves]) != leaves: raise ValueError('cannot compute node frequencies: the list of leaves differs between trees')

            # count if any set clade of provided tree matches with a clade of the reference tree
            desc = tree.map_descendants()
            for i in desc:
                if desc[i] in nodes: counts[nodes[desc[i]]] += 1

        # compute relative frequencies
        if relative:
            if n == 0: raise ValueError('cannot compute relative node frequencies: no trees provided')
            for node in counts: counts[node] = counts[node] / n

        # annotate nodes
        for node, count in counts.items(): node._label = count

    def clean_internal_labels(self):
        """
        Remove all internal node labels. This included the base of the tree.
        In practice, they are set to ``None``.
        """
        for node in self._nodes:
            if len(node._children) > 0: node._label = None

    def clean_branch_lengths(self):
        """
        Remove all branch lengths. In practice, they are set to
        ``None``.
        """
        for node in self._nodes:
            node._parent_branch = None # also the base 
            for child in node._children: node._child_branches[child] = None

    def remove_node(self, node, keep_parent=False):
        """
        Remove a node from the tree, as well as all its descendants.
        Since this operation may create a node with a single child, this
        method may remove the parent or the brother of the removed node
        depending on the structure of the tree,
        unless specified otherwise (see *keep_parent*).

        :param node: the node to remove, as a :class:`.Node` instance
            belowing to the current tree. Terminal nodes can be removed,
            but not the base of the tree.

        :param keep_parent: don't remove the parent of the
            removed node if it is left with only one child. If the
            parent is the base of the tree, keep the other descendant
            if it is not terminal (see below).

        Assume we remove node ``[3]`` from the tree with this
        structure::

                               /---------------------------->[2]
                               |
                 /----------->[1]           /--------------->[4]
                 |             |            |
                 |             \---------->[3]
                 |                          |
                [0]                         \--------------->[5]
                 |
                 |              /--------------------------->[7]
                 |              | 
                 \------------>[6]            /------------->[9]
                                |             |
                                \----------->[8]
                                              |
                                              \------------>[10]

        Then, we would end up with the following tree::

                 /----------->[1]--------------------------->[2]
                 |
                 |
                [0]
                 |              /--------------------------->[7]
                 |              | 
                 \------------>[6]            /------------->[9]
                                |             |
                                \----------->[8]
                                              |
                                              \------------>[10]

        The default behaviour is then to remove node ``[1]`` (and delete
        its label if it exists) and to set the length of the branch
        from ``[0]`` to ``[2]`` to the sum of the ``[0]`` to ``[1]`` and
        ``[1]`` to ``[2]``. But, with *keep_parent* is ``True``,
        the tree is left as is.

        There is a special case with the base of the tree. Assume that
        we remove node ``[1]`` from the original tree above. We then
        would have a non-standard structure with a single child to the
        base of the tree::

                                /--------------------------->[7]
                                | 
                [0]----------->[6]            /------------->[9]
                                |             |
                                \----------->[8]
                                              |
                                              \------------>[10]

        In that case, the base is not removed, but node ``[6]`` is
        removed using the :meth:`~.Tree.collapse` method (using option
        *ignore_len* set to ``False`` but *ignore_label* set to ``True``
        since the base is not supposed to bear a label). We end up with
        the following structure::

                 /--------------------------------->[7]
                 | 
                [0]                  /------------->[9]
                 |                   |
                 \----------------->[8]
                                     |
                                     \------------>[10]

        The length of the branch from ``[0]`` to ``[6]`` is spread
        equally between the branch from ``[0]`` to ``[7]`` and the
        branch from ``[0]`` to ``[8]`` (and so on if there are actually
        more than one descendants). If *keep_parent* is ``True`` or if
        ``[6]`` is a terminal node, it is not removed.
        """

        # sanity check
        if node not in self._nodes: raise ValueError('cannot delete a node that is not part of the tree')
        if node == self._base: raise ValueError('cannot delete the base of the tree')

        # record parent
        parent = node._parent

        # disconnect the subtree
        node._remove_parent()

        # remove nodes of the subtree
        for node in Tree._gather_nodes(node):
            node._unlink()
            self._nodes.remove(node)
            if node in self._leaves: self._leaves.remove(node)

        # collapse the ascendant/brother if needed
        if not keep_parent:
            if parent.num_children == 1:
                if parent == self._base:
                    brother = parent._children[0]
                    if brother.num_children != 0:
                        self.collapse(brother, ignore_len=False, ignore_label=True)
                else:
                    self.collapse(parent, ignore_len=False, ignore_label=True)

    @staticmethod
    def _gather_nodes(node):
        # collect all nodes from a subtree: return list of this node and
        # all its descendants
        nodes = [node]
        for child in node._children:
            nodes.extend(Tree._gather_nodes(child))
        return nodes

    def lateralize(self, reverse=False):
        """
        Flush bigger clades to one side of the tree.
        Modify the order of children of all nodes of the trees in such a
        way that they are sorted from the smallest to the
        largest number of descending leaves.

        :param reverse: sort from in the more-descendants to
            less-descendants order instead.
        """

        # get node mapping (only store number of descendants)
        nodes = self.map_descendants()
        for node in nodes: nodes[node] = len(nodes[node])

        # add leaves
        for leaf in self._leaves: nodes[leaf] = 1

        # sort all nodes of the tree
        for node in self._nodes:
            numdict = dict(zip(node._children, [nodes[child] for child in node._children]))
            node._children.sort(key=numdict.get, reverse=reverse)

    def midroot(self):
        """
        Automatic midpoint rooting of the tree. The tree must be
        initially unrooted (trifurcation at the root). This method
        identifies the most distant pair of terminal nodes (in case of
        a draw, one pair is picked randomly) and the root of the tree (as a
        new node) placed at the middle point of this path.
        """

        # sanity check
        if len(self._nodes) < 3: raise ValueError('cannot perform automatic rooting: not enough branches in tree')
        if self._base.num_children < 3: raise ValueError('cannot perform automatic rooting: tree must have a trifurcation at the root')

        # collect the path to root for all leaves
        paths = []
        for node in self._leaves:
            path = []
            while node.parent != None:
                L = node.parent_branch
                if L == None: raise ValueError('cannot perform automatic rooting: all branch lengths must be specified')
                path.append((node, L))
                node = node.parent
            paths.append(path)

        # find shortest path(s)
        best = []
        best_dist = - sys.float_info.max

        for i in range(len(paths)):
            for j in range(i+1, len(paths)):

                part1 = list(paths[i]) # deep copies to allow editing
                part2 = list(paths[j])
                while part1[-1] == part2[-1]: # the first of each at least is different
                    del part1[-1]
                    del part2[-1]
                path = [(node, L, False) for (node, L) in part1] + [(node, L, True) for (node, L) in part2[::-1]]
                d = sum([L for (node, L, reverse) in path])

                if (d - best_dist) > -0.000000001:
                    if (d - best_dist) > +0.000000001:
                        best_dist = d
                        best = []
                    best.append(path)

        # if more than several best, pick one randomly
        if len(best) > 1: best = best[random.irand(len(best))]
        else: best = best[0]

        # find midpoint
        lim = best_dist / 2
        acc = 0.0
        for node, L, reverse in best:
            acc += L
            if acc - lim > -0.000000001:
                outgroup = node
                if reverse: offset = min(1.0, (acc - lim) / L)
                else: offset = max(0.0, (acc - lim) / L)
                break
        else:
            raise RuntimeError('midpoint rooting: cannot find root location (please report bug)')

        # root itself
        self.root(outgroup, offset)

    def unroot(self, reverse=False):
        """
        Remove the root. The tree must be initially rooted
        (bifurcation at the root). This method removes the root node and
        places the base of the tree at one of the two basal nodes (the
        nodes that are ancestral to the two basal groups). This method
        does not change to total length of the tree. And error is raised
        if only one of the two basal branches has a length. If the
        initial basal node has a label, it is lost. If the node that
        becomes the base has a label, it is left there (it will appear
        a the base of the tree).

        :param reverse: if ``True``, place the base of the tree at the
            second basal node (by default, the first basal node is
            used).
        """
        if self._base.num_children != 2: raise ValueError('cannot unroot tree: tree is not rooted')
        node1 = self._base.child(0)
        node2 = self._base.child(1)
        L1 = node1.parent_branch
        L2 = node2.parent_branch
        if L1 == None and L2 == None: L = None
        elif L1 != None and L2 != None: L = L1 + L2
        else: raise ValueError('cannot unroot tree: only one of the two basal branches has a length')
        self._base._remove_child(node1)
        self._base._remove_child(node2)
        self._nodes.remove(self._base)
        if reverse:
            node2._connect(node1, L)
            self._base = node2
        else:
            node1._connect(node2, L)
            self._base = node1

class _tree_breath_first_iterator(object):
    def __init__(self, tree, start):
        self._this_level = [start]
        self._next_level = list(tree._base._children) # must be a deep copy

    def __iter__(self):
        return self

    def __next__(self):
        if len(self._this_level) == 0:
            if len(self._next_level) == 0:
                raise StopIteration
            else:
                self._this_level = self._next_level
                self._next_level = functools.reduce(list.__add__, [node._children for node in self._this_level])
        node = self._this_level.pop(0)
        return node

class _tree_depth_first_iterator(object):
    def __init__(self, tree, start):
        self._cur = start
        self._pile = []

    def __iter__(self):
        return self

    def __next__(self):
        if self._cur == None: raise StopIteration
        node = self._cur
        children = list(node.children())
        if len(children) == 0:
            if len(self._pile) == 0: self._cur = None
            else: self._cur = self._pile.pop()
        else:
            self._cur = children[0]
            self._pile.extend(children[1:])
        return node



"""
    Copyright 2015-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
from .. import eggwrapper as _eggwrapper, random
from .. import _interface, _tree, alphabets
from . import _param_helpers

class Simulator(object):
    """
    Manager of the coalescent simulator. The constructor takes arguments
    controlling the demographic and mutation models used for simulation.
    Only the number of populations is required at the time of
    construction. Once it is set, it can be never modified. Other
    constructor arguments are all optional and can be set or modified
    later using the :attr:`.params` instance attribute (either using its
    :meth:`~.ParamDict.update` method or the ``[]`` operator, such as in
    ``simulator.params['theta'] = 2.85``. Keyword arguments are passed as is to
    :meth:`~.ParamDict.update`. List-based parameters can be set if all
    values are provided in a sequence. Matrix-based parameters cannot be
    set here.

    :param num_pop: number of populations.
    :param migr: migration rate.

    Other keyword arguments can be any of the parameters defined in the
    table below.

    +-----------------+------------------------------------+-----------------+
    |Parameter        | Definition                         | Default         |
    +=================+====================================+=================+
    |``num_pop``      | Number of populations              | None, required  |
    +-----------------+------------------------------------+-----------------+
    |``num_sites``    | Number of sites                    | 0 (ISM)         |
    +-----------------+------------------------------------+-----------------+
    |``num_mut``      | Fixed number of mutations          | 0               |
    +-----------------+------------------------------------+-----------------+
    |``theta``        | :math:`4N_0\mu` parameter          | 0.0             |
    +-----------------+------------------------------------+-----------------+
    |``recomb``       | :math:`4N_0c` parameter            | 0.0             |
    +-----------------+------------------------------------+-----------------+
    |``mut_model``    | Mutation model (among ``KAM``,     | ``KAM``         |
    |                 | ``IAM``, ``SMM`` and ``TPM``)      |                 |
    +-----------------+------------------------------------+-----------------+
    |``TPM_proba``    | Probability parameter of TPM       | 0.5             |
    +-----------------+------------------------------------+-----------------+
    |``TPM_param``    | Shape parameter of TPM             | 0.5             |
    +-----------------+------------------------------------+-----------------+
    |``num_alleles``  | Number of alleles for KAM          | 2               |
    +-----------------+------------------------------------+-----------------+
    |``rand_start``   | Pick start allele randomly         | False           |
    |                 | for KAM (boolean)                  |                 |
    +-----------------+------------------------------------+-----------------+
    |``num_chrom``    | Number of sampled chromosomes      | 0 for all       |
    |                 | (per population)                   |                 |
    +-----------------+------------------------------------+-----------------+
    |``num_indiv``    | Number of sampled individuals      | 0 for all       |
    |                 | (per population)                   |                 |
    +-----------------+------------------------------------+-----------------+
    |``N``            | Population size, expressed         | 1.0 for all     |
    |                 | relatively to :math:`N_0`          |                 |
    |                 | (per population)                   |                 |
    +-----------------+------------------------------------+-----------------+
    |``G``            | Exponential growth/decline rate,   | 0.0 for all     |
    |                 | negative values mean decline       |                 |
    |                 | (per population)                   |                 |
    +-----------------+------------------------------------+-----------------+
    |``s``            | Population selfing probability     | 0.0 for all     |
    |                 | (per population)                   |                 |
    +-----------------+------------------------------------+-----------------+
    |``outgroup``     | Label of samples to place in the   | None            |
    |                 | outgroup (population index for     |                 |
    |                 | normal samples, label for delayed) |                 |
    +-----------------+------------------------------------+-----------------+
    |``site_pos``     | Site position, as values           | Equally spread  |
    |                 | between 0.0 and 1.0                |                 |
    |                 | (per site)                         |                 |
    +-----------------+------------------------------------+-----------------+
    |``site_weight``  | Site mutation weight, controlling  | 1.0 for all     |
    |                 | the relative probability of sites  |                 |
    |                 | (per site)                         |                 |
    +-----------------+------------------------------------+-----------------+
    |``migr_matrix``  | Pairwise migration rate matrix     | 0.0 for all     |
    |                 | (the diagonal cannot be set)       |                 |
    +-----------------+------------------------------------+-----------------+
    |``trans_matrix`` | Matrix of transition weights       | 1.0 for all     |
    |                 | between pairs of alleles           |                 |
    +-----------------+------------------------------------+-----------------+
    |``events``       | List of events added by the user   | Empty           |
    +-----------------+------------------------------------+-----------------+
    |``max_iter``     | Maximum number of iterations       | 100,000         |
    +-----------------+------------------------------------+-----------------+

    The following table presents the categories of events that can be
    added to the ``events`` list using :meth:`~.EventList.add`.

    +------------------+-------------------------+----------------------------------+
    |Event code        | Description             | Parameters                       |
    +==================+=========================+==================================+
    |``size``          | Change population size  | ``T`` -- date (1)                |
    |                  |                         +----------------------------------+
    |                  |                         | ``N`` -- new size                |
    |                  |                         +----------------------------------+
    |                  |                         | ``idx`` -- population index (2)  |
    +------------------+-------------------------+----------------------------------+
    |``migr``          | Change all migration    | ``T`` -- date (1)                |
    |                  | rate                    +----------------------------------+
    |                  |                         | ``M`` -- migration rate (all     |
    |                  |                         | pairwise migration rates are     |
    |                  |                         | set to ``M/(num_pop-1)``)        |
    +------------------+-------------------------+----------------------------------+
    |``pair_migr``     | Change pairwise         | ``T`` -- date (1)                |
    |                  | migration rate          +----------------------------------+
    |                  |                         | ``M`` -- pairwise migration      |
    |                  |                         | rate                             |
    |                  |                         +----------------------------------+
    |                  |                         | ``src`` -- source population     |
    |                  |                         | index                            |
    |                  |                         +----------------------------------+
    |                  |                         | ``dst`` -- destination           |
    |                  |                         | population index                 |
    +------------------+-------------------------+----------------------------------+
    |``growth``        | Change population       | ``T`` -- date (1)                |
    |                  | exponential             +----------------------------------+
    |                  | growth/decline rate     | ``G`` -- new rate                |
    |                  |                         +----------------------------------+
    |                  |                         | ``idx`` -- population index (2)  |
    +------------------+-------------------------+----------------------------------+
    |``selfing``       | Change population       | ``T`` -- date (1)                |
    |                  | self-fertilization      +----------------------------------+
    |                  | rate                    | ``s`` -- new rate                |
    |                  |                         +----------------------------------+
    |                  |                         | ``idx`` -- population index (2)  |
    +------------------+-------------------------+----------------------------------+
    |``recombination`` | Change recombination    | ``T`` -- date (1)                |
    |                  | rate                    +----------------------------------+
    |                  |                         | ``R`` -- new rate                |
    +------------------+-------------------------+----------------------------------+
    |``bottleneck``    | Apply a bottleneck      | ``T`` -- date (1)                |
    |                  |                         +----------------------------------+
    |                  |                         | ``S`` -- bottleneck strength (3) |
    |                  |                         +----------------------------------+
    |                  |                         | ``idx`` -- population index (2)  |
    +------------------+-------------------------+----------------------------------+
    |``admixture``     | Move lineages from one  | ``T`` -- date (1)                |
    |                  | population to another   +----------------------------------+
    |                  |                         | ``proba`` -- migration           |
    |                  |                         | probability (in [0,1] range      |
    |                  |                         +----------------------------------+
    |                  |                         | ``src`` -- source population     |
    |                  |                         | index                            |
    |                  |                         +----------------------------------+
    |                  |                         | ``dst`` -- destination           |
    |                  |                         | population index                 |
    +------------------+-------------------------+----------------------------------+
    |``merge``         | Merge a population to   | ``T`` -- date (1)                |
    |                  | another (take all       +----------------------------------+
    |                  | lineages from ``src``,  | ``src`` -- source population     |
    |                  | move them to ``dst``,   | index                            |
    |                  | and remove ``src``      +----------------------------------+
    |                  |                         | ``dst`` -- destination           |
    |                  |                         | population index                 |
    +------------------+-------------------------+----------------------------------+
    |``sample``        | Perform a delayed       | ``T`` -- date (1)                |
    |                  | a some point in the     +----------------------------------+
    |                  | past in one of the      | ``idx`` -- population index      |
    |                  | populations             |                                  |
    |                  |                         +----------------------------------+
    |                  |                         | ``label`` -- group label (4)     |
    |                  |                         +----------------------------------+
    |                  |                         | ``num_chrom`` -- number of       |
    |                  |                         | sampled chromosomes              |
    |                  |                         +----------------------------------+
    |                  |                         | ``num_indiv`` -- number of       |
    |                  |                         | sampled individuals              |
    +------------------+-------------------------+----------------------------------+

    1. Time is expressed in units of :math:`4N_0` generations.
    2. If ``idx`` is omitted, the event is applied to all populations 
       at once.
    3. Bottleneck strength is expressed in time units. A bottleneck is
       implemented as a period of time during which coalescences are the
       only event allowed to occur.
    4. The label of delayed sample can be set to the same value than the
       populations index (in this case, delayed samples will have the
       same label than normal samples from the same population), or to a
       different label, as the user's option. Based on the value of the
       *outgroup* option, this label controls whether delayed samples
       should go to the outgroup.

    The parameters ``num_chrom``, ``num_indiv``, ``N``, ``G``, ``s``,
    ``site_pos``, and ``site_weight`` are represented by
    :class:`.ParamList` instances that behave like lists (except that
    their length cannot be changed). In particular, :class:`.ParamList`
    support subscript indexing and it can also be initialized by passing
    a sequence. The parameters ``migr_matrix`` and ``trans_matrix`` are
    represented by :class:`.ParamMatrix` instances that support a
    double-index subscript system to read/changes values (as in
    ``params['migr_matrix'][i,j]`` to access the value at row *i* and
    column *j*. Diagonal values are read as ``None`` and cannot be
    changed. Finally, ``events`` is represented by a :class:`.EventList`
    instance that exhibit limited list functionality and provides
    methods to add, read, and modify events. See this class for more
    information about out editing the list of events.
    """

    def __init__(self, num_pop, migr=0.0, **kwargs):
        self._alphabet_unltd = alphabets.Alphabet('range', (-_eggwrapper.MAX_ALLELE_RANGE, _eggwrapper.MAX_ALLELE_RANGE+1), (0, 0), name='unlimited')
        self._alphabet_positive = alphabets.Alphabet('range', (0, _eggwrapper.MAX_ALLELE_RANGE), (0, 0), name='positive unlimited')
        self._alphabets_K = {}
        self._params = _param_helpers.ParamDict(num_pop)
        self._params.set_migr(migr)
        self._params.update(**kwargs)
        self._coalesce = _eggwrapper.Coalesce()
        self._align = _interface.Align._create_from_data_holder(self._coalesce.data(), self._alphabet_unltd)

    @property
    def params(self):
        """
        Simulation parameters of this instance. This object is a
        instance of :class:`.ParamDict`, which is a clone
        of :class:`dict` that does not let the users add or remove
        parameters, but lets them modify values of parameters
        (similarly, the number of items of per-population or per-site
        parameters cannot be modified).
        """
        return self._params

    def _reset_align(self):
        self._align._ns = self._params._params.get_nsam()
        if self._params['mut_model'] == 'KAM':
            K = self._params['num_alleles']
            if K not in self._alphabets_K:
                self._alphabets_K[K] = alphabets.Alphabet('int', range(K), [], name='KAM:{0}'.format(K))
            alph = self._alphabets_K[K]
        elif self._params['mut_model'] == 'IAM': alph = self._alphabet_positive
        else: alph = self._alphabet_unltd
        self._align._alphabet = alph

    def simul(self, dest=None):
        """
        Perform a single simulation. The simulation is conditioned on
        the current value of parameters stored in :attr:`.params`.

        :param dest: An :class:`.Align` to reset using simulated data. All
            previous data will be lost.

        :return: A new :class:`.Align` instance containing the simulated
            data unless *dest* is specified (otherwise ``None``).

        .. note::

            The method :meth:`.iter_simul` can be much more efficient.
            For performance-critical applications, its use is
            recommended.
        """

        # run the simulation
        self._coalesce.simul(self._params._params, True)

        # reset the local align instance
        self._reset_align()

        # return a deep copy of the alignment
        if dest is None: return _interface.Align.create(_interface.Align._create_from_data_holder(self._coalesce.data(), self._align._alphabet))
        else: dest._reset_from_data_holder(self._coalesce.data())

    def iter_simul(self, nrepet, cs=None, dest_trees=None, **feed_params):

        """
        Perform several simulations. Simulations are conditioned on the
        current value of parameters
        stored in :attr:`.params`. Return an iterator that will loop
        over the requested number of simulations. The simulated
        alignments are always available at each iteration loop as the
        instance attribute :attr:`.align`. By default, all iterations
        return a reference to this instance, but if *cs* is specified a
        dictionary of statistics is returned at each simulation.

        :param nrepet: number of simulations. If ``None``, iterate for
            ever (you are required to use ``break`` in loop with your
            own stopping criterion is reach).
        :param cs: :class:`.ComputeStats` instance properly configured
            to compute statistics from simulated data. If *cs* is specified,
            each iteration round will yield the dictionary of statistics
            obtained from :meth:`~.ComputeStats.process_align` called on
            this object. See also  *cs_filter* and *cs_struct*.
        :param dest_trees: a :class:`list` in which simulated trees will
            be appended. Since each simulation can yield several trees
            (because of recombination), a new sub-list will be appended
            for each simulation, with one item for each tree. Each tree
            covered a defined region of the simulated chromosome, so
            each item of each sub-list will be ``(tree, start, stop)``
            :class:`!tuple` with ``tree`` as a new :class:`.Tree` instance, and
            ``start`` and ``stop`` as the start and stop positions (real
            numbers comprised between 0 and 1). If recombination
            occurred, trees will not be sorted within their sub-list. If
            recombination did not occur, each simulation will be
            represented by a list with a single :class:`!tuple`. Any previously
            data contained in *dest_tree* is left untouched.
        :param feed_params: other arguments provide sequences of values
            for any parameter (except ``num_pop``),
            allowing to modify any set of parameters between
            simulations. All changes are permanent and affect any later
            simulation. All additional options must be in the
            ``key=value`` format, with have one of the parameters as
            ``key``, and a sequence of values as ``value``. All
            sequences must be of length at least equal to *nrepet*. If
            longer, additional values are ignored. For list-based
            parameters, each value must be a ``(index, value)`` tuple
            and, for matrix-based parameters, each value must be a
            ``(index1, index2, value)`` tuple.

        :return: An iterator, yielding a reference to :attr:`.align` if
            *cs* is ``None``, or otherwise a dictionary of computed
            statistics.
        """

        # reset the local align instance
        self._reset_align()

        # safety checking
        for key, values in feed_params.items():
            if key not in self._params._keys: raise ValueError('invalid parameter: `{0}`'.format(key))
            if len(values) < nrepet: raise ValueError('not enough values provided for: `{0}`'.format(key))

        # main loop
        i = 0
        stop = -1 if nrepet is None else nrepet
        while i != nrepet:

            # set the provided params
            for key, values in feed_params.items():
                value = values[i]
                self._params._set(key, value)

            # run the simulation
            self._coalesce.simul(self._params._params, False)

            # export tree if needed
            if dest_trees != None:
                dest_trees.append([
                    ( _tree.Tree._from_coalesce(self._coalesce, i, str), 
                      self._coalesce.tree_start(i),
                      self._coalesce.tree_stop(i) )
                        for i in range(self._coalesce.number_of_trees()) ])

            # mutate
            self._coalesce.mutate()

            # return what is requested
            if cs != None: yield cs.process_align(self._align)
            else: yield self._align

            i += 1

        # reset object
        self._align.reset()

    @property
    def align(self):
        """
        An :class:`.Align` instance containing simulated data for the
        current iteration of :meth:`.iter_simul`. The data in this
        instance will be updated at each iteration round, and deleted at
        the end of the iteration. If it must be copied, a deep copy is
        required (typically with :meth:`.Align.create`).
        """
        return self._align



"""
    Copyright 2008-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import re, sys, functools
from io import StringIO
from ..tools import _seq_manip

class GenBankFeatureLocation(object):
    """
    Hold the location of a GenBank feature. Supports various forms of
    location specifications as allowed in the GenBank format. The
    constructor allows to parse a GenBank-formatted
    string. By default, features are on the forward strand and segmented
    features are ranges (not orders).

    In addition to methods documented below, the following operations
    are supported for ``loc`` if it is a
    :class:`.GenBankFeatureLocation` instance:

    +--------------------------------+---------------------------------------------+
    |Operation                       | Result                                      |
    +================================+=============================================+
    |``len(loc)``                    | Number of segments                          |
    +--------------------------------+---------------------------------------------+
    |``loc[index]``                  | Return the ``(first, last)`` :class:`tuple` |
    +--------------------------------+---------------------------------------------+
    |``for (first, last) in loc:``   | Iterator over segments                      |
    +--------------------------------+---------------------------------------------+
    |``params.format()``             | Generate a GenBank representation           |
    +--------------------------------+---------------------------------------------+
    """

    def __init__(self, string=None):
        self._complement = False # True or False
        self._range = True # True or False
        self._type = '' # string of S|R|B|C for Single, Range, Between and Choice
        self._pos = [] # list of start/stop positions, always in increasing number
        self._ends = [] # list of boolean tuples (True if the segment is 5'/3' partial)
        if string!=None:
            self._parse(string)

    def __iter__(self):
        for pos in self._pos: yield pos

    def __len__(self):
        return len(self._pos)

    def __getitem__(self, index):
        return self._pos[index]

    def __str__(self):
        return self.format()

    def format(self):
        """
        String representation of the instance.
        """
        if not len(self._pos):
            raise RuntimeError('cannot format Genbank feature\'s location: no positions were loaded')
        string = []
        for i in range(len(self._pos)):
            if self._type[i]=='S':
                string.append(str(self._pos[i][0] + 1)) # the final string will be unicode even on py2
            elif self._type[i]=='R' or self._type[i]=='C':
                item = ''
                if self._ends[i][0]:
                    item += '<'
                item += str(self._pos[i][0] + 1)
                item += '.'
                if self._type[i]=='R':
                    item +='.'
                if self._ends[i][1]:
                    item += '>'
                item += str(self._pos[i][1] + 1)
                string.append(item)
            elif self._type[i]=='B':
                string.append(  '%d^%d' %(self._pos[i][0] +1 , self._pos[i][1] +1) )
            else: raise RuntimeError('internal error in GenBankFeatureLocation.format()')
        string = ','.join(string)
        if len(self._pos)>1:
            if self._range:
                string = 'join(%s)' %string
            else:
                string = 'order(%s)' %string
        if self._complement:
            string = 'complement(%s)' %string
        return string

    def copy(self):
        """
        Deep copy of the current instance.
        """
        gbfl = GenBankFeatureLocation()
        gbfl._complement = self._complement
        gbfl._range = self._range
        gbfl._type = str(self._type)
        gbfl._pos = [(i,j) for i,j in self._pos]
        gbfl._ends = [(i,j) for i,j in self._ends]
        return gbfl

    def _parse(self, string):

        # process top-level location groups

        # complement not incompatible with the rest
        if string[:11]=='complement(':
            if string[-1]!=')':
                raise IOError('invalid GenBank feature location: %s' %string)
            self._complement = True
            string = string[11:-1]

        # range or order (only remove it)
        if string[:5] == 'join(':
            if string[-1]!=')':
                raise IOError('invalid GenBank feature location: %s' %string)
            string = string[5:-1]
        elif string[:6] == 'order(':
            if string[-1]!=')':
                raise IOError('invalid GenBank feature location: %s' %string)
            self._range = True
            string = string[6:-1]

        # extracts the segment(s)
        segments = string.split(',')
        for segment in segments:

            # position formats
            single = re.match('^(\d+)$', segment)
            rangeChoice   = re.match('^(\<?)(\d+)(\<?)(\.\.?)(\>?)(\d+)(\>?)$', segment)
            between = re.match('^(\d+)\^(\d+)$', segment)

            if single:
                try:
                    pos = int(single.group(1))  -1
                except ValueError:
                    raise IOError('invalid position in GenBankFeature location: %s' %segment)
                self.add_single_base(pos)

            elif rangeChoice:

                # gets positions
                try:
                    start = int(rangeChoice.group(2))  -1
                    stop = int(rangeChoice.group(6))   -1
                except ValueError:
                    raise IOError('invalid position in GenBankFeature location: %s' %segment)

                # gets partial marks
                Ls = rangeChoice.group(1)+rangeChoice.group(3)
                if Ls=='':
                    Lp = False
                elif Ls=='<':
                    Lp = True
                else:
                    raise IOError('invalid specification in GenBankFeature location: %s' %segment)
                Rs = rangeChoice.group(5)+rangeChoice.group(7)
                if Rs=='':
                    Rp = False
                elif Rs=='>':
                    Rp = True
                else:
                    raise IOError('invalid specification in GenBankFeature location: %s' %segment)
    
                # loads appropriate feature
                if rangeChoice.group(4)=='..':
                    self.add_base_range(start, stop, Lp, Rp)
                elif rangeChoice.group(4)=='.':
                    self.add_base_choice(start, stop, Lp, Rp)
                else:
                    raise RuntimeError('error in GenBankFeatureLocation constructor')

            elif between:
                try:
                    start = int(between.group(1))  -1
                    stop = int(between.group(2))   -1
                except ValueError:
                    raise IOError('invalid position in GenBankFeature location: %s' %segment)
                if stop!=start+1:
                    raise IOError('invalid between-base feature specification: %s' %segment)
                self.add_between_base(start)
            
            else:
                raise IOError('invalid GenBank feature segment positions: %s' %segment)

    def set_complement(self):
        """
        Place the feature on the complement strand.
        """
        self._complement = True

    def set_forward(self):
        """
        Place the feature on the forward strand. This
        is the default.
        """
        self._complement = False

    def is_complement(self):
        """
        ``True`` if the feature is on the complement strand.
        """
        return self._complement

    def as_order(self):
        """
        Define the feature as an order instead of a range.
        """
        self._range = False

    def as_range(self):
        """
        Define the feature as a range. This is the default.
        """
        self._range = True

    def is_range(self):
        """
        ``True`` if the feature is a range. ``False`` if
        it is an order.
        """
        return self._range

    def shift(self, offset):
        """
        Shift all positions. The argument can be positive or negative.
        """
        self._pos = list(map(lambda x: [x[0]+offset, x[1]+offset], self._pos))

    def add_single_base(self, position):
        """
        Add a single-base segment to the feature.
        """
        if not isinstance(position, int):
            raise TypeError('GenBankFeatureLocation positions must be of type int')
        if len(self._pos) and position<self._pos[-1][1]:
            raise ValueError('GenBankFeatureLocation positions must be entered in increasing order')

        self._pos.append((position,position))
        self._type += 'S'
        self._ends.append((False,False))

    def add_between_base(self, position):
        """
        Add a segment lying between two consecutive  bases.
        The feature will be set
        between *position* and *position* + 1. If the feature is
        intended to be placed on the complement strand between
        positions, say, 1127 and 1128, one must use
        ``add_between_base(1127)`` in combination with
        :meth:`~.io.GenBankFeatureLocation.set_complement`.
        """
        if not isinstance(position, int):
            raise TypeError('GenBankFeatureLocation positions must be of type int')
        if len(self._pos) and position<self._pos[-1][1]:
            raise ValueError('GenBankFeatureLocation positions must be entered in increasing order')

        self._pos.append((position,position+1))
        self._type += 'B'
        self._ends.append((False,False))

    def add_base_range(self, first, last, left_partial=False, right_partial=False):
        """
        Add a base range to the feature.

        :param first: first position of the range.
        :param last: last position of the range (included).
        :param first_partial: specify that the real start of the segment
            is somewhere 5' of *first*.
        :param first_partial: specify that the real end of the segment
            is somewhere 3' of *last*.

        If the feature is intended to
        be placed on the complement strand between positions, say, 1127
        and 1482, one must use ``add_base_range(1127,1482)`` in
        combination with :meth:`~.io.GenBankFeatureLocation.set_complement`.
        """
        if not isinstance(first, int) or not isinstance(last, int):
            raise TypeError('GenBankFeatureLocation positions must be of type int')
        if len(self._pos) and first<self._pos[-1][1]:
            raise ValueError('GenBankFeatureLocation positions must be entered in increasing order')
        if (last<first):
            raise ValueError('GenBankFeatureLocation positions must be entered in increasing order')
        self._pos.append((first,last))
        self._type += 'R'
        self._ends.append(( bool(left_partial), bool(right_partial) ))

    def add_base_choice(self, first, last, left_partial=False, right_partial=False):
        """
        Segment corresponding to a single base in a given range. Arguments
        are identical to :meth:`~.io.GenBankFeatureLocation.add_base_range`.
        """
        if not isinstance(first, int) or not isinstance(last, int):
            raise TypeError('GenBankFeatureLocation positions must be of type int')
        if len(self._pos) and first<self._pos[-1][1]:
            raise ValueError('GenBankFeatureLocation positions must be entered in increasing order')
        if (last<first):
            raise ValueError('GenBankFeatureLocation positions must be entered in increasing order')
        if (first==last):
            raise ValueError('invalid use of GenBankFeatureLocation.add_base_choice: start=stop (use add_single_base instead)')
        self._pos.append((first,last))
        self._type += 'C'
        self._ends.append(( bool(left_partial), bool(right_partial) ))

    def rc(self, length):
        """
        Reverse the feature positions. Positions are modified to be
        counted from the end.

        :param length: length of the complete sequence (required).
        """
        self._complement = not self._complement
        self._pos = self._pos[::-1]
        self._ends = self._ends[::-1]
        self._type = self._type[::-1]
        for i in range(len(self._pos)):
            a,b = self._pos[i]
            self._pos[i] = (length-b-1, length-a-1)
            self._ends[i] = ( self._ends[i][1], self._ends[i][0] )

class GenBankFeature(object):
    """
    Feature of a GenBank record.
    :class:`.io.GenBankFeature` instances must be only used along an
    :class:`.io.GenBank` instance. The constructor creates an empty
    instance (although a parent :class:`.io.GenBank` instance is required)
    and either :meth:`~.io.GenBankFeature.update` or :meth:`~.io.GenBankFeature.parse` must be used
    subsequently.

    :param parent: an :class:`.io.GenBank` instance to which the feature
        should be attached.

    Note that ``str(feature)`` is equivalent to ``feature.format()``.
    """

    def __init__(self, parent):
        self._parent = parent
        self._type = ''
        self._location = GenBankFeatureLocation()
        self._qualifiers = []

    def get_type(self):
        """
        Type of the instance.
        """
        return self._type

    def qualifiers(self):
        """
        Dictionary with all qualifier values. This method
        cannot be used to change data within the instance.
        """
        return dict(self._qualifiers)

    def add_qualifier(self, key, value):
        """
        Add a qualifier.
        """
        self._qualifiers.append((key, value))

    def update(self, feat_type, location, ** qualifiers):
        """
        Update feature information.

        :param feat_type: a string identifying the feature type (such as
            ``"gene"``, ``"CDS"``, ``"misc_feature"``, etc.). All
            strings are acceppted.
        :param location: an :class:`.io.GenBankFeatureLocation` instance
            giving the feature's location.
        :param qualifiers: other qualifiers must be passed as keyword
            arguments. It is not allowed to use ``"type"`` as a
            qualifier keyword.
        """
        if 'type' in qualifiers:
            raise ValueError('cannot use "type" as custom qualifier in `GenBankFeature`\'s constructor')
        self._qualifiers = [(i,qualifiers[i]) for i in qualifiers]
        self._location = location
        self._type = feat_type

    def parse(self, string):
        """
        Update feature information from a string.

        :param string: a GenBank-formatted string.
        """
        try:
            self._type = string.split()[0]
            locstring = string.split()[1]
            string = string.split()[2:]
        except IndexError:
            raise IOError('invalid GenBank feature string')
        while len(string) and string[0][0]!='/':
            locstring += string.pop(0)
        qualifiers=string

        # now we have the feature's position in a genuine string (locstring)
        self._location = GenBankFeatureLocation(locstring)

        # and the rest of qualifiers in a list (qualifiers)
        self._qualifiers = []
        if not len(qualifiers):
            return

        # fuses qualifiers where needed
        i=1
        while i<len(qualifiers):
            if qualifiers[i][0]!='/' or '=' not in qualifiers[i]:
                qualifiers[i-1]+=' '+qualifiers[i]
                del qualifiers[i]
            else:
                i+=1
        
        # processes the qualifiers
        for qualifier in qualifiers:
            if qualifier[0]!='/':
                raise IOError('invalid qualifier string: %s' %qualifier)
            try:
                pos = qualifier.index('=')
            except ValueError:
                raise IOError('invalid qualifier string: %s' %qualifier)
            key = qualifier[1:pos]
            value = qualifier[pos+1:]
            if key=='translation':
                value = ''.join(value.split())

            self._qualifiers.append((key,value))

    def get_sequence(self):
        """
        Return the string corresponding to this feature. If the
        positions pass beyond the end of the parent's sequence, a
        :exc:`RuntimeError` (and not an
        :exc:`IndexError`) is raised.
        """
        seq = ''
        for i,j in self._location:
            if j>len(self._parent):
                raise RuntimeError('GenBank feature exceeds sequence length')
            seq += self._parent._sequence[i:j+1]
        if self._location.is_complement(): return _seq_manip.rc(seq)
        else: return seq

    def get_start(self):
        """
        First position of the first segment.
        """
        return self._location[0][0]

    def get_stop(self):
        """
        Stop position of the last segment. 
        """
        return self._location[-1][1]

    def copy(self, genbank):
        """
        Return a copy of the current instance.

        :param genbank: :class:`GenBank` instance to which the returned
            instance should be attached.
        """
        feature = GenBankFeature(genbank)
        feature.update(
            self._type, self._location.copy(), **dict(self._qualifiers))
        return feature

    def shift(self, offset):
        """
        Shift all positions. The argument can be positive or negative.
        """
        self._location.shift(offset)

    def __str__(self):
        return self.format()

    def format(self):
        """
        GenBank-formatted string representing the feature.
        """
        string = '     %s %s\n'%(self._type.ljust(15), str(self._location))
        string = self._parent._wrap(string, 21) + '\n'
        for key, value in self._qualifiers:
            if '/' in value:
                if value[0]+value[-1]!='""' and value[0]+value[-1]!='\'\'':
                    value = '"%s"' %value
            string+= self._parent._wrap('                     /%s=%s' %(key,value), 21)
            string+= '\n'
        return string

    def rc(self, length=None):
        """
        Reverse-complement the feature.

        :param length: length of the complete sequence (by default, take
            the information directly from the parent).
        """
        if length==None:
            length = len(self._parent)
        self._location.rc(length)

class GenBank(object):
    """
    Process a GenBank-formatted DNA sequence record.

    :param fname: input file name.
    :param string: GenBank-formatted string.

    Only one of the two arguments *fname* and *string* can be
    non-``None``. If both are ``None``, the constructor generates an
    empty instance with a sequence of length 0. If *fname* is
    non-``None``, a GenBank record is read from the file with this name.
    If *string* is non-``None``, a GenBank record is read
    from this string. The following variables are read from the parsed
    input if present: *accession*, *definition*, *title*, *version*,
    *GI*, *keywords*, *source*, *references* (which is a list), *locus*
    and *others*. Their default value is ``None`` except for
    *references* and *others* for which default is an empty list.
    *source* is a (*description*, *species*, *taxonomy*) :class:`!tuple`. Each of
    *references* is a (*header*, *raw reference*) :class:`!tuple` and each of
    *others* is a (*key*, *raw*) :class:`!tuple`.

    In addition to methods documented below, the following operations
    are supported for ``gb`` if it is a :class:`.GenBank` instance:

    +--------------------+-------------------------------------------------+
    |Operation           | Result                                          |
    +====================+=================================================+
    |``len(gb)``         | length of the sequence attached to this record  |
    +--------------------+-------------------------------------------------+
    |``for feat in gb:`` | Iterate over :class:`.GenBankFeature` instances |
    |                    | of this record                                  |
    +--------------------+-------------------------------------------------+
    |``str(gb)``         | GenBank representation of the record            |
    +--------------------+-------------------------------------------------+
    """

    def __init__(self, fname=None, string=None):
        self._sequence = ''
        self._features = []
        self.accession = None
        self.definition = None
        self.title = None
        self.locus = None
        self.version = None
        self.GI = None
        self.source = (None,None,None)
        self.references = []
        self.keywords = None
        self.others = []
        if fname!=None and string!=None:
            raise ValueError('GenBank constructor expects at most one argument')
        stream = None
        if fname:
            stream = open(fname)
        if string:
            stream = StringIO(string)
        if stream:
            self._parse(stream)

    def add_feature(self, feature):
        """
        Add a feature to the instance.

        :param feature: an :class:`.io.GenBankFeature` instance.
        """
        self._features.append(feature)

    def number_of_features(self):
        """
        Number of features contained in the instance.
        """
        return len(self._features)

    @property
    def sequence(self):
        """
        Sequence string. Note that changing the
        record's string might invalidate the features (meaning that the
        setting an invalid sequence might cause the features to point to
        incorrect or out-of-bounds regions of the sequence).
        """
        return self._sequence

    @sequence.setter
    def sequence(self, string): self._sequence = string

    def __iter__(self):
        for feature in self._features:
            yield feature

    def extract(self, from_pos, to_pos):
        """
        Extract a subset of the instance.

        :param from_pos: start position.
        :param to_pos: stop position (not included).
        :return: A new :class:`.io.GenBank` instance representing a subset of
          the current instance. All features that are completely included in the
          specified range are exported.
        """
        if from_pos < 0 or to_pos>=len(self):
            raise ValueError('invalid positions for GenBank extraction')
        gb = GenBank()
        gb._sequence = self._sequence[from_pos:to_pos]
        for feature in self:
            if feature.get_start() >= from_pos and feature.get_stop() < to_pos:
                clone = feature.copy(gb)
                clone.shift(-from_pos)
                gb._features.append( clone )
        return gb

    def _parse(self, stream):
        
        # identifies blocks marked by a capitalized word at the very
        # beginning of the line, and send them to the dynamic parser
        
        key = None  # current block key
        block = [] # block being read
        
        while True:

            line = stream.readline()
            if not len(line): break

            if line.strip()=='ORIGIN':
                match = re.match('(.+)()', line.strip())
            else:
                match = re.match('^([A-Z]+) (.+)', line)
            if match:
                if key:
                    self._parse_block(key, block)
                key = match.group(1)
                block = [ match.group(2) ]
                if key=='ORIGIN':
                    break
            else:
                block.append(line)
                
        if key!='ORIGIN':
            raise IOError('GenBank records lacks sequence')
        
        self._parse_sequence(stream)

    def _parse_block(self, key, content):
        
        # alias "the dynamic parser". Processes the different GenBank
        # blocks appropriately
        
        def merge(items):
            return ' '.join([' '.join(i.split()) for i in items])

        if key=='LOCUS':
            if len(content)>1:
                raise IOError('GenBank record exhibits an invalid LOCUS line')
            self.locus = content[0].strip()

        elif key=='DEFINITION':
            self.definition = merge(content)

        elif key=='ACCESSION':
            self.accession = merge(content)

        elif key=='VERSION':
            content = merge(content)
            match = re.match('([^ ]+) +GI\:(\d+)', content)
            if not match:
                raise IOError('incorrect VERSION/GI line in GenBank record')
            self.version = match.group(1)
            self.GI = match.group(2)

        elif key=='KEYWORDS':
            self.keywords = merge(content)

        elif key=='SOURCE':
            if len(content)<3:
                raise IOError('incorrect SOURCE section in GenBank record')
            else:
                source= content[0].strip()
                organism= content[1].strip()
                if organism[:8] != 'ORGANISM':
                    raise IOError('incorrect SOURCE section in GenBank record')
                organism= organism[10:]
                taxonomy= merge(content[2:])
            self.source = source, organism, taxonomy

        elif key=='REFERENCE':
            if len(content)<2:
                raise IOError('incorrect REFERENCE section in GenBank record')
            self.references.append((content[0].strip(), ''.join(content[1:])))

        elif key=='FEATURES':
            self._parse_features(content[1:])

        else:
            self.others.append(( key, content[0]+'\n'+''.join(content[1:]).rstrip() ))

    def _parse_features(self, lines):
        buff = ''
        for line in lines:
            # new feature
            if not re.match('^ {8}', line):
                if len(buff):
                    feature = GenBankFeature(self)
                    feature.parse(buff)
                    self._features.append(feature)
                    buff=''
            buff+=line
        if len(buff):
            feature = GenBankFeature(self)
            feature.parse(buff)
            self._features.append(feature)

    def _parse_sequence(self, stream):
        self._sequence = ''
        while True:
            line= stream.readline()
            if not len(line):
                raise IOError('GenBank sequence doesn\'t terminated appropriately')
            line=line.strip()
            match = re.match(' *[0-9]+ ([A-Za-z ]+)', line)
            if match:
                fragment = ''.join( match.group(1).split())
                self._sequence += fragment.upper()
            elif line=='//':
                break
            else:
                raise IOError('invalid sequence line in GenBank record (reproduced below)\n%s' %line)

    def __len__(self):
        return len(self._sequence)

    def write(self, fname):
        """
        Write the formatted record to a file.
        """
        f = open(fname, 'w')
        try:
            self.write_stream(f)
        finally:
            f.close()

    def __str__(self):
        return self.format()

    def format(self):
        """
        String representation of the instance.
        """
        stream = StringIO()
        self.write_stream(stream)
        return stream.getvalue()

    def write_stream(self, stream):
        """
        From the formatted record to a stream.

        :param stream: a file-compatible object.
        """
        try: stream.write('LOCUS       %s\n' %self.locus)
        except AttributeError: raise TypeError('`stream` must be a file-type object')
        if self.definition!=None:
            stream.write(self._wrap('DEFINITION  %s' %self.definition, 12) + '\n')
        if self.accession!=None:
            stream.write('ACCESSION   %s\n' %self.accession)
        if self.version!=None or self.GI!=None:
            stream.write('VERSION     %s  GI:%s\n' %(self.version, self.GI))
        if self.keywords!=None:
            stream.write(self._wrap('KEYWORDS    %s' %self.keywords, 12)+ '\n')
        if (self.source[0]!=None or self.source[1]!=None or self.source[2]!=None):
            stream.write(self._wrap('SOURCE      %s' %str(self.source[0]), 12)+ '\n')
        if self.accession!=None:
            stream.write(self._wrap('  ORGANISM  %s' %str(self.source[1]), 12)+ '\n')
            stream.write(self._wrap('            %s' %str(self.source[2]), 12)+ '\n')

        for reference in self.references:
            stream.write('REFERENCE   %s\n%s' %reference)

        for other in self.others:
            # ignores "BASE COUNT"
            if other[0]=='BASE' and other[1].split()[0]=='COUNT':
                continue
            # otherwise writes
            stream.write('%s %s\n' %other)    

        stream.write('FEATURES             Location/Qualifiers\n')
        for feature in self:
            stream.write(feature.format())

        stream.write('ORIGIN')
        
        if len(self) > 999999999:
            raise IOError('cannot export a GenBank instance with a sequence longest than 999999999 base pairs')

        c=0
        while c<len(self._sequence):
            stream.write('\n' +str(c+1).rjust(9))
            for i in range(6):
                stream.write(' '+self._sequence[c:c+10].lower())
                c+=10
                if c>=len(self._sequence):
                    break
        stream.write('\n//\n')

    _LINEWIDTH = 80
    _MAXBREAK = 40

    def _wrap(self, string, spacing):
        cache = ''.join([i for i in string])
        space=0
        comma=0
        c=0
        res = ''
        while c<len(cache):
            if cache[c]==' ':
                space=c
            if cache[c]==',':
                comma=c
            c+=1
            if c==(self._LINEWIDTH-1):
                if c<len(cache) and cache[c]==' ':
                    space = c
                if space>=self._MAXBREAK:
                    res+=cache[:space]
                    res+= '\n'
                    cache = ''.join([' ']*spacing) + cache[space+1:]
                    c=0
                    space=0
                    comma=0
                elif comma>=self._MAXBREAK:
                    res+=cache[:comma+1]
                    res+= '\n'
                    cache = ''.join([' ']*spacing) + cache[comma+1:]
                    c=0
                    space=0
                    comma=0
                else:
                    res+=cache[:c]
                    res+= '\n'
                    cache = ''.join([' ']*spacing) + cache[c:]
                    c=0
                    comma=0
                    space=0
        res += cache
        return res.rstrip()

    def rc(self):
        """
        Reverse-complement the instance (in place). All features
        positions and the sequence will be reverted and applied to the
        complementary strand. The features will be sorted in increasing
        start position (after reverting). This method should be applied
        only on genuine nucleotide sequences.
        """
        self._sequence= _seq_manip.rc(self._sequence)
        for feature in self: feature.rc(len(self))
        self._features.sort(key=GenBankFeature.get_start)



"""
    Copyright 2018-2023 Stephane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
from io import StringIO
from .. import _interface

def to_ms(aln, outfile, positions=None, alphabet=None, as_codes=False,
    converter=None, spacer='auto',  header='//\n'):
    """
    Export data in the ``ms`` program format. The program is described in
    Hudson (2002 [*Bioinformatics* **18**: 337-338]).
    The format is as follows:

    * One header line with two slashes (can be extended/replaced).
    * One line with the number of sites
    * One line with the positions, only if the number of sites is larger
      than zero.
    * The matrix of genotypes (one line per sample), only if the number
      of sites is larger than zero.
    * One empty line.

    ``ms`` generates binary data (0/1). This function may export any
    integer values depending on the data contained in the input object.

    :param data: :class:`.Align` instance.

    :param outfile: where to write data. This argument can be (i) a file
        (or file-like) object open for writing, (ii) the name of a file
        (which will be created) as a string, or (iii) `None` (in which
        case the resulting string will be returned).

    :param positions: list of site positions, with length matching
        the alignment length. Positions are required to be in the [0,1]
        range but the order is not checked. By default (if the argument
        value is ``None``), sites are supposed to be evenly spread over
        the [0,1] interval.

    :param alphabet: an :class:`.Alphabet` instance to use
        instead of the alphabet attached to the input alignment. When
        using this argument, alleles are transformed using the alphabet
        provided as argument using a direct mapping of the input alignment
        alphabet to the provided alphabet.

    :param as_codes: use the internal alphabet codes for
        exporting. Useful if a non-int alphabet is used. Ignored if
        *converter* is used. Not allowed if *alphabet* is used.

    :param converter: a user-provided function generating the appropriate
        allelic value based on an input alphabet code. The simple inline
        example below shows how to convert any missing data to 0 and shift all
        valid alleles of one unit: ``converter=lambda x: 0 if x<0 else x+1``.

    :param spacer: insert a space between all genotypes
        (to separate each locus/site). By default, all genotypes are
        concatenated as in the standard ``ms`` format.
        In this case, it is required that all allelic
        values (or codes if *as_codes* is specified) are >=0 and <10. By
        default (``"auto"``), the spacer is automatically inserted if
        the alphabet defines alleles or codes outside the [0, 9] range,
        or if *converter* is specified.

    :param header: string to print instead of the two slashes line.
        Must contain the final new line.

    :return: A ``ms``-formatted string if *outfile* is ``None``,
        otherwise ``None``.
    """

    # error checking
    if not isinstance(aln, _interface.Align): raise TypeError('an Align instance is expected')
    ls = aln.ls
    if positions is not None:
        if len(positions) != ls: raise ValueError('invalid number of items in the positions list')
        if min(positions) < 0.0 or max(positions) > 1.0: raise ValueError('out of range item in the positions list')
    if alphabet is None: alph = aln.alphabet
    else:
        if as_codes:
            raise ValueError('options `as_codes` and `alphabet` are incompatible')
        alph = alphabet
    if alph.type not in ['int', 'range'] and converter is None and as_codes == False: raise ValueError('unsupported alphabet type: {0}'.format(alph.type))
    if converter is not None and as_codes:
        raise ValueError('options `as_codes` and `converter` are incompatible')

    # manage output file
    if outfile is None:
        out = StringIO()
    elif isinstance(outfile, str): out = open(outfile, 'w')
    else: out = outfile
    out.write(header)

    # shows number of sites and site
    out.write('segsites: {0}\n'.format(ls))
    if positions is None:
        if ls == 0: positions = []
        elif ls == 1: positions = [0.5]
        else: positions = [i / (ls-1) for i in range(ls)]
    if ls > 0:
        out.write('positions: {0}\n'.format(' '.join(['{0:.5f}'.format(i) for i in positions])))
    else:
        out.write('\n')

    # spacer
    if spacer == 'auto':
        if converter is not None: spacer = True
        else:
            if as_codes:
                if alph._obj.num_missing() > 0 or alph._obj.num_exploitable() > 10: spacer = True
                else: spacer = False
            else:
                if alph.type == 'range':
                    if alph._obj.min_value() < 0 or alph._obj.max_value() > 9: spacer = True
                    else: spacer = False
                else:
                    exp, mis = alph.get_alleles()
                    alls = exp + mis
                    if min(alls) < 0 or max(alls) > 9: spacer = True
                    else: spacer = False

    # export data
    if converter is not None: f = converter
    elif as_codes: f = lambda x: x
    else: f = alph.get_value
    for i in range(aln.ns):
        values = tuple(map(f, [aln._obj.get_sample(i, j) for j in range(ls)]))
        if not spacer and (min(values) < 0 or max(values) > 9): raise ValueError('not printable allele for ms format (you should insert a spacer)')
        if spacer: out.write(' '.join(map(str, values)) + '\n')
        else: out.write(''.join(map(str, values)) + '\n')

    # close file or return string as necessary
    out.write('\n')
    if isinstance(outfile, str): out.close()
    elif outfile is None: return out.getvalue()



"""
    Copyright 2023 Thomas Coudoux, Stéphane De Mita, Mathieu Siol

    This file is part of EggLib.

    EggLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EggLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with EggLib.  If not, see <http://www.gnu.org/licenses/>.
"""

import os, egglib, sys, unittest, tempfile
from collections.abc import Iterable

class Structure_test(unittest.TestCase):
    def test_get_samples(self):
        d = {
            '1000': { # cluster
                '1100': {    # pop
                    '1110': (0, 1),  # indiv
                    '1120': (2, 3),
                    '1130': (4, 6),
                    '1140': (8, 10)
                },
                '1200': {
                    '1210': (15, 16),
                    '1230': (17, 18),
                    '1240': (19, 20)
                }
            },
            '2000': {
                '2100': {
                    '2110': (21, 22),
                    '2120': (23, 24),
                    '2130': (25, 27),
                    '2140': (26, 28)
                },
                    '2200': {
                    '2210': (29, 30),
                    '2220': (31, 32),
                    '2230': (33, 34)
                }
            }
        }
        do = {
            '910': (55, 56),
            '920': (57, 12)
        }

        s = egglib.struct_from_dict(d, do)
        self.assertSetEqual(s.get_samples(),
            set([0,1,2,3,4,6,8,10,15,16,17,18,19,20,21,22,23,24,25,27,26,28,29,30,31,32,33,34]))

    def test_subsetting(self):
        coal = egglib.coalesce.Simulator(4, num_chrom=[5, 2, 5, 4], theta=5.0, migr=0.2)
        cs = egglib.stats.ComputeStats()
        cs.add_stats('nseff', 'S', 'thetaW', 'Pi', 'D', 'FstWC')
        for aln in coal.iter_simul(200):
            sub = aln.subset([5, 6, 12, 13, 14, 15])
            cs.configure(struct=egglib.struct_from_labels(sub, lvl_pop=0))
            ctrl = cs.process_align(sub)
            struct = egglib.struct_from_labels(aln, lvl_pop=0)
            struct = struct.as_dict()[0]
            del struct[None]['0']
            del struct[None]['2']
            struct = egglib.struct_from_dict(struct, None)
            cs.configure(struct=struct)
            test = cs.process_align(aln)
            self.assertDictEqual(ctrl, test)
            if ctrl['S'] > 0:
                self.assertEqual(ctrl['nseff'], 6)

    def test_subset(self):
        struct = egglib.struct_from_dict(
            {'clu1': {
                'pop1': {
                    'idv1': (0, 1), 'idv2': (2, 3), 'idv3': (4, 5), 'idv4': (6, 7)},
                'pop2': {
                    'idv5': (8, 9), 'idv6': (10, 11), 'idv7': (12, 13), 'idv8': (14, 15)},
                'pop3': {
                    'idv9': (16, 17), 'idv10': (18, 19), 'idv11': (20, 21), 'idv12': (22, 23)}
                },
            'clu2': {
                'pop4': {
                    'idv13': (24, 25), 'idv14': (26, 27), 'idv15': (28, 29), 'idv6': (30, 31)},
                'pop5': {
                    'idv17': (32, 33), 'idv18': (34, 35), 'idv19': (36, 37), 'idv20': (38, 39)},
                'pop6': {
                    'idv21': (40, 41), 'idv22': (42, 43), 'idv23': (44, 45), 'idv24': (46, 47)},
                'pop7': {
                    'idv25': (48, 49), 'idv26': (50, 51), 'idv27': (52, 53), 'idv28': (54, 55)}
                }}, 
                    {'otg1': (56, 57), 'otg2': (58, 59), 'otg3': (60, 61), 'otg4': (62, 63)}
        )

        site = [0, 0, 0, 0, 0, 0, 0, 0, # pop1
                1, 1, 1, 1, 2, 2, 2, 2, # pop2
                3, 3, 3, 3, 3, 3, 3, 3, # pop3
                3, 3, 3, 3, 3, 3, 3, 3, # pop4
                4, 4, 4, 5, 5, 5, 6, 6, # pop5
                7, 7, 7, 7, 7, 7, 7, 7, # pop6
                7, 7, 7, 7, 8, 8, 8, 8, # pop7
                8, 8, 8, 8, 9, 9, 9, 9] # otg

        site = egglib.site_from_list(site, alphabet=egglib.alphabets.positive_infinite)

        cs = egglib.stats.ComputeStats(struct=struct)
        cs.add_stats('Aing', 'ns_site', 'Atot', 'Aotg')
        stats = cs.process_site(site)
        self.assertEqual(stats['Aing'], 9)
        self.assertEqual(stats['Aotg'], 2)
        self.assertEqual(stats['Atot'], 10)

        cs.configure(struct=None)
        self.assertEqual(cs.process_site(site)['Aing'], 10)

        self.assertRaisesRegex(ValueError, 'there must be at least one population', struct.subset)
        self.assertRaisesRegex(ValueError, 'invalid population label: prout', struct.subset, ['prout'])

        cs.configure(struct=struct.subset(['pop1', 'pop5']))
        stats = cs.process_site(site)
        self.assertEqual(stats['Aing'], 4)
        self.assertEqual(stats['Aotg'], 2)
        self.assertEqual(stats['Atot'], 6)

        cs.configure(struct=struct.subset(['pop5'], outgroup=False))
        stats = cs.process_site(site)
        self.assertEqual(stats['Aing'], 3)
        self.assertIsNone(stats['Aotg'])
        self.assertEqual(stats['Atot'], 3)

        cs.configure(struct=struct.subset(['pop1', 'pop6', 'pop6', 'pop7'], outgroup=True))
        stats = cs.process_site(site)
        self.assertEqual(stats['Aing'], 3)
        self.assertEqual(stats['Aotg'], 2)
        self.assertEqual(stats['Atot'], 4)

        cs.configure(struct=struct.subset(pops=['pop5'], clusters=['clu1'], outgroup=True))
        stats = cs.process_site(site)
        self.assertEqual(stats['Aing'], 7)
        self.assertEqual(stats['Aotg'], 2)
        self.assertEqual(stats['Atot'], 9)


    def test_shuffle(self):
        struct = {}
        struct['c1'] = {}
        struct['c1']['p1'] = {}
        struct['c1']['p1']['i1'] = (0, 1)
        struct['c1']['p1']['i2'] = (2, 3)
        struct['c1']['p1']['i3'] = (4, 5)
        struct['c1']['p2'] = {}
        struct['c1']['p2']['i4'] = (6, 7)
        struct['c1']['p2']['i5'] = (8, 9)
        struct['c1']['p2']['i6'] = (10, 11)
        struct['c1']['p2']['i7'] = (12, 13)
        struct['c1']['p3'] = {}
        struct['c1']['p3']['i8'] = (14, 15)
        struct['c1']['p3']['i9'] = (16, 17)
        struct['c1']['p3']['i10'] = (18, 19)
        struct['c2'] = {}
        struct['c2']['p4'] = {}
        struct['c2']['p4']['i11'] = (20, 21)
        struct['c2']['p4']['i12'] = (22, 23)
        struct['c2']['p4']['i13'] = (24, 25)
        struct['c2']['p4']['i14'] = (26, 27)
        struct['c2']['p5'] = {}
        struct['c2']['p5']['i15'] = (28, 29)
        struct['c2']['p5']['i16'] = (30, 31)
        struct['c2']['p5']['i17'] = (32, 33)
        struct['c2']['p5']['i18'] = (34, 35)
        struct = egglib.struct_from_dict(struct, {'i19': (36, 37)})

        def f(x): return int(x[1:])
        def show(s):
            ret = [[],[]]
            ing, otg = struct.as_dict()
            for c in sorted(ing, key=f):
                for p in sorted(ing[c], key=f):
                    for i in sorted(ing[c][p], key=f):
                        ret[0].extend(ing[c][p][i])
            for i in sorted(otg, key=f):
                ret[1].extend(otg[i])
            return ret

        # original structure fingerprint
        original = show(struct)

        # default mode (single shuffling)
        for i in 'it', 'ic', 'is', 'st', 'sc', 'ct':
            n = 0
            for rep in range(20):
                with struct.shuffle(i):
                    ret = show(struct)
                    if ret[0] != original[0]: n += 1
                    self.assertListEqual(ret[1], original[1])
            self.assertGreater(n, 5)
            ret = show(struct)
            self.assertListEqual(ret[0], original[0])
            self.assertListEqual(ret[1], original[1])

        # show that iteration not possible in default mode
        n = 0
        for i in range(10):
            with struct.shuffle() as shuffler:
                ret = show(struct)
                if ret[0] != original[0]: n += 1
                assert ret[1] == original[1]
                with self.assertRaises(TypeError) as cm:
                    for i in shuffler: pass # should not be iterable
                self.assertIn('is not iterable', str(cm.exception))
        self.assertGreater(n, 5) # check at least 5 times different
        ret = show(struct)
        self.assertListEqual(ret[0], original[0])
        self.assertListEqual(ret[1], original[1])

        # test iteration
        with struct.shuffle(nr=100) as shuffler:
            c = 0
            n = 0
            for i in shuffler:
                self.assertEqual(i, c)
                ret = show(struct)
                if ret[0] != original[0]: n += 1
                self.assertListEqual(ret[1], original[1])
                c += 1
            self.assertEqual(c, 100)
            self.assertGreater(n, 5) # check at least 5 times different
        ret = show(struct)
        self.assertListEqual(ret[0], original[0])
        self.assertListEqual(ret[1], original[1])

    def test_labels(self):

        # check that empty labels are forbidden
        fas = """\
>sam1@pop1
AAAAAAAAAA
>sam2@pop1
AAAAAAAAAA
>sam3@pop1,pop2,pop3
AAAAAAAAAA
>sam4
AAAAAAAAAA
>sam5@pop1,
AAAAAAAAAA
>sam6@pop1
AAAAAAAAAA
>sam7@pop1,,pop3
AAAAAAAAAA
>sam8@pop1
AAAAAAAAAA
"""

        with self.assertRaises(IOError):
            aln = egglib.io.from_fasta_string(fas, alphabet=egglib.alphabets.DNA, labels=True)

        # repair the fasta
        fas = fas.replace(',\n', '\n')
        fas = fas.replace(',,', ',pop2,')

        # import repaired fasta
        aln = egglib.io.from_fasta_string(fas, alphabet=egglib.alphabets.DNA, labels=True)
        check = [ ('sam1', ['pop1']),
                  ('sam2', ['pop1']),
                  ('sam3', ['pop1','pop2','pop3']),
                  ('sam4', []),
                  ('sam5', ['pop1']),
                  ('sam6', ['pop1']),
                  ('sam7', ['pop1','pop2','pop3']),
                  ('sam8', ['pop1'])]
        self.assertListEqual([(seq.name, list(seq.labels)) for seq in aln], check)

        # attempt to set 0-length label
        with self.assertRaises(ValueError):
            aln[0].labels.append('')

        with self.assertRaises(ValueError):
            aln[0].labels[0] = ''

        # test None (automatic if level not specified / allowed in structure input only if only one item)
        d = {None: {None: {'i1': [0, 1], 'i2': [2, 3], 'i3': [6, 7], 'i4': [8, 9]}}}, {'i1': [10, 11]}
        struct = egglib.struct_from_dict(*d)
        dx = struct.as_dict()
        self.assertDictEqual(dx[0], d[0])
        self.assertDictEqual(dx[1], d[1])

        d = {'c1': {}, 'c2': {None: {'i1': [0, 1], 'i2': [2, 3], 'i3': [6, 7], 'i4': [8, 9]}}}, {'i1': [10, 11]}
        with self.assertRaises(ValueError):
            struct = egglib.struct_from_dict(*d)

        d = ({'c1': {None: {'i1': [0, 1], 'i2': [2, 3]}}}, {'i1': [4,5]})
        with self.assertRaises(ValueError):
            struct = egglib.struct_from_dict(*d)

        # check that non-represented levels are automatically set to a single None item
        aln = egglib.io.from_fasta_string("""\
>sam1@pop1
AAAAAAAAAA
>sam2@pop1
AAAAAAAAAA
>sam3@pop2
AAAAAAAAAA
>sam4@pop2
AAAAAAAAAA
>sam5@#
AAAAAAAAAA
>sam6@#
AAAAAAAAAA
""", alphabet=egglib.alphabets.DNA, labels=True)

        dx = egglib.struct_from_labels(aln, lvl_pop=0).as_dict()
        self.assertDictEqual(dx[0], {None: {'pop1': {'0': [0], '1': [1]}, 'pop2': {'2': [2], '3': [3]}}})
        self.assertDictEqual(dx[1], {'4': [4], '5': [5]})

        aln = egglib.io.from_fasta_string("""\
>sam1@i1
AAAAAAAAAA
>sam2@i1
AAAAAAAAAA
>sam3@i2
AAAAAAAAAA
>sam4@i2
AAAAAAAAAA
>sam5@#,i1
AAAAAAAAAA
>sam6@#,i1
AAAAAAAAAA
""", alphabet=egglib.alphabets.DNA, labels=True)

        dx = egglib.struct_from_labels(aln, lvl_indiv=0).as_dict()
        self.assertDictEqual(dx[0], {None: {None: {'i1': [0, 1], 'i2': [2, 3]}}})
        self.assertDictEqual(dx[1], {'i1': [4,5]})

        aln = egglib.io.from_fasta_string("""\
>sam1@c1,i1
AAAAAAAAAA
>sam2@c1,i1
AAAAAAAAAA
>sam3@c1,i2
AAAAAAAAAA
>sam4@c1,i2
AAAAAAAAAA
>sam5@#,i1
AAAAAAAAAA
>sam6@#,i1
AAAAAAAAAA
""", alphabet=egglib.alphabets.DNA, labels=True)

        dx = egglib.struct_from_labels(aln, lvl_clust=0, lvl_indiv=1).as_dict()
        self.assertDictEqual(dx[0], {'c1': {'c1': {'i1': [0, 1], 'i2': [2, 3]}}})
        self.assertDictEqual(dx[1], {'i1': [4,5]})

        # check outgroup sample
        dx = egglib.struct_from_labels(aln, lvl_clust=0, lvl_indiv=1, skip_outgroup=True).as_dict()
        self.assertDictEqual(dx[0], {'c1': {'c1': {'i1': [0, 1], 'i2': [2, 3]}}})
        self.assertDictEqual(dx[1], {})

        # support missing samples
        aln = egglib.io.from_fasta_string("""\
>sam1@idv1,pop1
AAAAAAAAAA
>sam2@idv2,pop1
AAAAAAAAAA
>sam3@idv3,pop1
AAAAAAAAAA
>sam4
AAAAAAAAAA
>sam5@idv4
AAAAAAAAAA
>sam6@idv5,pop2
AAAAAAAAAA
>sam7@idv6,pop2
AAAAAAAAAA
>sam8@#
AAAAAAAAAA
""", alphabet=egglib.alphabets.DNA, labels=True)

        dx = egglib.struct_from_labels(aln, lvl_indiv=0, lvl_pop=1).as_dict()
        self.assertDictEqual(dx[0], {None: {'pop1': {'idv1': [0], 'idv2': [1], 'idv3': [2]}, 'pop2': {'idv5': [5], 'idv6': [6]}}})
        self.assertDictEqual(dx[1], {})

        aln = egglib.io.from_fasta_string("""\
>sam1@idv1,pop1
AAAAAAAAAA
>sam2@idv2,pop1
AAAAAAAAAA
>sam3@idv3,pop1
AAAAAAAAAA
>sam4
AAAAAAAAAA
>sam5@idv4
AAAAAAAAAA
>sam6@idv5,pop2
AAAAAAAAAA
>sam7@idv6,pop2
AAAAAAAAAA
>sam8@#,idv1
AAAAAAAAAA
""", alphabet=egglib.alphabets.DNA, labels=True)

        dx = egglib.struct_from_labels(aln, lvl_indiv=0, lvl_pop=1).as_dict()
        self.assertDictEqual(dx[0], {None: {'pop1': {'idv1': [0], 'idv2': [1], 'idv3': [2]}, 'pop2': {'idv5': [5], 'idv6': [6]}}})
        self.assertDictEqual(dx[1], {'idv1': [7]})

    def test_outgroup_label(self):

        # structure with 2 pops + 1 `outgroup` single-indiv pop
        coal = egglib.coalesce.Simulator(3, num_indiv=[5, 5, 1], migr_matrix=[[None, 1, 0], [1, None, 0], [0, 0, None]])
        coal.params.add_event('merge', T=3, src=2, dst=0)
        coal.params.add_event('merge', T=3, src=1, dst=0)

        # perform a simulation
        aln = coal.simul()

        # shows the labels
        self.assertListEqual(list(aln[0].labels), ['0', '0'])
        self.assertListEqual(list(aln[1].labels), ['0', '0'])
        self.assertListEqual(list(aln[2].labels), ['0', '1'])
        self.assertListEqual(list(aln[3].labels), ['0', '1'])
        self.assertListEqual(list(aln[4].labels), ['0', '2'])
        self.assertListEqual(list(aln[5].labels), ['0', '2'])
        self.assertListEqual(list(aln[6].labels), ['0', '3'])
        self.assertListEqual(list(aln[7].labels), ['0', '3'])
        self.assertListEqual(list(aln[8].labels), ['0', '4'])
        self.assertListEqual(list(aln[9].labels), ['0', '4'])
        self.assertListEqual(list(aln[10].labels), ['1', '5'])
        self.assertListEqual(list(aln[11].labels), ['1', '5'])
        self.assertListEqual(list(aln[12].labels), ['1', '6'])
        self.assertListEqual(list(aln[13].labels), ['1', '6'])
        self.assertListEqual(list(aln[14].labels), ['1', '7'])
        self.assertListEqual(list(aln[15].labels), ['1', '7'])
        self.assertListEqual(list(aln[16].labels), ['1', '8'])
        self.assertListEqual(list(aln[17].labels), ['1', '8'])
        self.assertListEqual(list(aln[18].labels), ['1', '9'])
        self.assertListEqual(list(aln[19].labels), ['1', '9'])
        self.assertListEqual(list(aln[20].labels), ['2', '10'])
        self.assertListEqual(list(aln[21].labels), ['2', '10'])

        # make structure with three populations ignoring individual level
        ing, otg = egglib.struct_from_labels(aln, lvl_pop=0).as_dict()
        self.assertDictEqual(ing, {None: {
            '0': {    '0': [0],  '1': [1],   '2': [2],   '3': [3],   '4': [4],   '5': [5],   '6': [6],   '7': [7],   '8': [8],   '9': [9]},
            '1': {  '10': [10], '11': [11], '12': [12], '13': [13], '14': [14], '15': [15], '16': [16], '17': [17], '18': [18], '19': [19]},
            '2': {  '20': [20], '21': [21]}}})
        self.assertDictEqual(otg, {})

        # make structure with three populations
        ing, otg = egglib.struct_from_labels(aln, lvl_pop=0, lvl_indiv=1).as_dict()
        self.assertDictEqual(ing, {None: {
            '0': {  '0': [ 0, 1], '1': [ 2, 3], '2': [ 4, 5], '3': [ 6, 7], '4': [ 8, 9]},
            '1': {  '5': [10,11], '6': [12,13], '7': [14,15], '8': [16,17], '9': [18,19]},
            '2': { '10': [20,21]}}})
        self.assertDictEqual(otg, {})

        # two populations + outgroup ignoring individual level
        ing, otg = egglib.struct_from_labels(aln, lvl_pop=0, outgroup_label='2').as_dict()
        self.assertDictEqual(ing, {None: {
            '0': {    '0': [0],  '1': [1],   '2': [2],   '3': [3],   '4': [4],   '5': [5],   '6': [6],   '7': [7],   '8': [8],   '9': [9]},
            '1': {  '10': [10], '11': [11], '12': [12], '13': [13], '14': [14], '15': [15], '16': [16], '17': [17], '18': [18], '19': [19]}}})
        self.assertDictEqual(otg, {'20': [20], '21': [21]})

        # two populations + outgroup
        ing, otg = egglib.struct_from_labels(aln, lvl_pop=0, lvl_indiv=1, outgroup_label='2').as_dict()
        self.assertDictEqual(ing, {None: {
            '0': {  '0': [ 0, 1], '1': [ 2, 3], '2': [ 4, 5], '3': [ 6, 7], '4': [ 8, 9]},
            '1': {  '5': [10,11], '6': [12,13], '7': [14,15], '8': [16,17], '9': [18,19]}}})
        self.assertDictEqual(otg, {'10': [20,21]})

    def test_from_samplesizes(self):
        struct = egglib.struct_from_samplesizes([5, 5], ploidy=2, outgroup=1)
        ref = ({None:
                {'pop1': {'idv1': [0, 1], 'idv2': [2, 3], 'idv3': [4, 5],
                          'idv4': [6, 7], 'idv5': [8, 9]},
                 'pop2': {'idv6': [10, 11], 'idv7': [12, 13],
                          'idv8': [14, 15], 'idv9': [16, 17],
                          'idv10': [18, 19]}}}, {'idv11': [20, 21]})
        self.assertTupleEqual(struct.as_dict(), ref)

        self.assertTupleEqual(egglib.struct_from_samplesizes([6], ploidy=1, outgroup=0).as_dict(), ({
            None: {'pop1': {'idv1': [0], 'idv2': [1], 'idv3': [2], 'idv4': [3], 
                    'idv5': [4], 'idv6': [5]}}}, {}))

        self.assertTupleEqual(egglib.struct_from_samplesizes([], ploidy=4, outgroup=2).as_dict(), ({
            None: {}}, {'idv1': [0, 1, 2, 3], 'idv2': [4, 5, 6, 7]}))

        self.assertTupleEqual(egglib.struct_from_samplesizes([4, 1, 0, 2], ploidy=3, outgroup=2).as_dict(), (
            {None: {
                'pop1': {'idv1': [0, 1, 2], 'idv2': [3, 4, 5], 'idv3': [6, 7, 8], 'idv4': [9, 10, 11]},
                'pop2': {'idv5': [12, 13, 14]},
                'pop3': {},
                'pop4': {'idv6': [15, 16, 17], 'idv7': [18, 19, 20]}}},
            {'idv8': [21, 22, 23], 'idv9': [24, 25, 26]}))

        self.assertTupleEqual(egglib.struct_from_samplesizes([0], ploidy=1, outgroup=0).as_dict(), ({None: {'pop1': {}}}, {}))

        self.assertTupleEqual(egglib.struct_from_samplesizes([], ploidy=1, outgroup=0).as_dict(), ({None: {}}, {}))

    def test_from_iterable(self):

        ### development test ###
        aln = egglib.Align(alphabet=egglib.alphabets.DNA)
        aln.add_samples([
            ('name1', 'AAAAAA', ['pop1', 'idv1']),
            ('name2', 'AAAAAA', ['pop1', 'idv1']),
            ('name3', 'AAAAAA', ['pop1', 'idv2']),
            ('name4', 'AAAAAA', ['pop1', 'idv2']),
            ('name5', 'AAAAAA', ['pop2', 'idv3']),
            ('name6', 'AAAAAA', ['pop2', 'idv3']),
            ('name7', 'AAAAAA', ['pop2', 'idv4']),
            ('name8', 'AAAAAA', ['pop2', 'idv4']),
            ('nameA', 'AAAAAA', ['#', 'idv5']),
            ('nameB', 'AAAAAA', ['#', 'idv5'])])


        struct1 = egglib.struct_from_labels(aln, lvl_pop=0, lvl_indiv=1)
        assert struct1.as_dict() == (
            {None: {'pop1': {'idv1': [0, 1], 'idv2': [2, 3]},
                    'pop2': {'idv3': [4, 5], 'idv4': [6, 7]}}},
                                                    {'idv5': [8, 9]})

        labels = ['pop1'] * 4 + ['pop2'] * 4 + [None] * 2
        struct2 = egglib.struct_from_iterable(labels)
        assert struct2.as_dict() == (
            {None: {'pop1': {'idv1': [0], 'idv2': [1], 'idv3': [2], 'idv4': [3]},
                    'pop2': {'idv5': [4], 'idv6': [5], 'idv7': [6], 'idv8': [7]}}}, {})


        labels = [['pop1'], ['pop1'], ['pop1'], ['pop1'], ['pop2'], ['pop2'], ['pop2'], ['pop2'], [None], [None]]
        struct3 = egglib.struct_from_iterable(labels, fmt='P')
        assert struct3.as_dict() == (
            {None: {'pop1': {'idv1': [0], 'idv2': [1], 'idv3': [2], 'idv4': [3]},
                    'pop2': {'idv5': [4], 'idv6': [5], 'idv7': [6], 'idv8': [7]}}}, {})

        labels = [['name1', 'pop1'], ['name2', 'pop1'], ['name3', 'pop1'],
                  ['name4', 'pop1'], ['name5', 'pop2'], ['name6', 'pop2'],
                  ['name7', 'pop2'], ['name8', 'pop2'], ['nameA', None], ['nameB', None]]
        struct4 = egglib.struct_from_iterable(labels, fmt='NP', data=aln)
        assert struct4.as_dict() == (
            {None: {'pop1': {'name1': [0], 'name2': [1], 'name3': [2], 'name4': [3]},
                    'pop2': {'name5': [4], 'name6': [5], 'name7': [6], 'name8': [7]}}}, {})

        ### iterable ###
        # pass a list
        labels = ['pop1', 'pop1', 'pop1', 'pop1', 'pop2', 'pop2', 'pop2', 'pop2']
        ctrl = {None: {'pop1': {'idv1': [0], 'idv2': [1], 'idv3': [2], 'idv4': [3]},
                       'pop2': {'idv5': [4], 'idv6': [5], 'idv7': [6], 'idv8': [7]}}}
        struct = egglib.struct_from_iterable(labels)
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # pass a file/map
        try:
            f, fname = tempfile.mkstemp()
            os.write(f, b'\n'.join(map(str.encode, labels)))
            os.close(f)
            with open(fname) as f:
                struct = egglib.struct_from_iterable(f, function=str.strip)
            ing, otg = struct.as_dict()
            self.assertDictEqual(ing, ctrl)
            self.assertDictEqual(otg, {})
        finally:
            if os.path.isfile(fname): os.unlink(fname)

        # pass an invalid type
        with self.assertRaises(TypeError):
            error = egglib.struct_from_iterable(142)

        with self.assertRaises(TypeError):
            error = egglib.struct_from_iterable('strings not supported')

        # empty string
        labels[3] = ''
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels)

        # pass a list if a string is expected
        labels = [['pop1'], ['pop1'], ['pop1'], ['pop1'], ['pop2'], ['pop2'], ['pop2'], ['pop2']]
        with self.assertRaises(TypeError):
            error = egglib.struct_from_iterable(labels, fmt=None)

        # opposite
        labels = ['pop1', 'pop1', 'pop1', 'pop1', 'pop2', 'pop2', 'pop2', 'pop2']
        with self.assertRaises(TypeError):
            error = egglib.struct_from_iterable(labels, fmt='P')

        ### fmt ###
        # just P
        labels = [['pop1'], ['pop1'], ['pop1'], ['pop1'], ['pop2'], ['pop2'], ['pop2'], ['pop2']]
        ctrl = {None: {'pop1': {'idv1': [0], 'idv2': [1], 'idv3': [2], 'idv4': [3]},
                       'pop2': {'idv5': [4], 'idv6': [5], 'idv7': [6], 'idv8': [7]}}}
        struct = egglib.struct_from_iterable(labels, fmt='P')
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # just C (error)
        labels = [['clu1'], ['clu1'], ['clu1'], ['clu1'], ['clu2'], ['clu2'], ['clu2'], ['clu2']]
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='C')

        # C and I (error)
        labels = [['clu1', 'idv1'],
                  ['clu1', 'idv1'],
                  ['clu1', 'idv2'],
                  ['clu1', 'idv2'],
                  ['clu2', 'idv3'],
                  ['clu2', 'idv3'],
                  ['clu2', 'idv4'],
                  ['clu2', 'idv4']]
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='CI')

        # C and P
        labels = [['clu1', 'pop1'],
                  ['clu1', 'pop1'],
                  ['clu1', 'pop1'],
                  ['clu1', 'pop1'],
                  ['clu2', 'pop2'],
                  ['clu2', 'pop2'],
                  ['clu2', 'pop3'],
                  ['clu2', 'pop3']]
        ctrl = {'clu1': {'pop1': {'idv1': [0], 'idv2': [1], 'idv3': [2], 'idv4': [3]}},
                'clu2': {'pop2': {'idv5': [4], 'idv6': [5]},
                         'pop3': {'idv7': [6], 'idv8': [7]}}}
        struct = egglib.struct_from_iterable(labels, fmt='CP')
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # P and I
        labels = [['p1', '1'],
                  ['p1', '1'],
                  ['p1', '2'],
                  ['p1', '2'],
                  ['p2', '3'],
                  ['p2', '3'],
                  ['p2', '4'],
                  ['p2', '4']]
        ctrl = {None: {'p1': {'1': [0, 1], '2': [2, 3]},
                       'p2': {'3': [4, 5], '4': [6, 7]}}}
        struct = egglib.struct_from_iterable(labels, fmt='PI')
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # I
        labels = [['north'], ['north'], ['north'], ['west'], ['west'], ['west'], ['south'], ['south'], ['south'], ['east'], ['east'], ['east']]
        ctrl = {None: {None: {'north': [0, 1, 2], 'west': [3, 4, 5], 'south': [6, 7, 8], 'east': [9, 10, 11]}}}
        struct = egglib.struct_from_iterable(labels, fmt='I')
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # CPI
        labels = [['x1', 'clu1', 'pop1'],
                  ['x2', 'clu1', 'pop1'],
                  ['x3', 'clu1', 'pop1'],
                  ['x4', 'clu1', 'pop1'],
                  ['x5', 'clu2', 'pop2'],
                  ['x6', 'clu2', 'pop2'],
                  ['x7', 'clu2', 'pop3'],
                  ['x8', 'clu2', 'pop3']]
        ctrl = {'clu1': {'pop1': {'x1': [0], 'x2': [1], 'x3': [2], 'x4': [3]}},
                'clu2': {'pop2': {'x5': [4], 'x6': [5]},
                         'pop3': {'x7': [6], 'x8': [7]}}}
        struct = egglib.struct_from_iterable(labels, fmt='ICP')
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # invalid ploidy
        labels = [['x1', 'clu1', 'pop1'],
                  ['x1', 'clu1', 'pop1'],
                  ['x3', 'clu1', 'pop1'],
                  ['x4', 'clu1', 'pop1'],
                  ['x5', 'clu2', 'pop2'],
                  ['x6', 'clu2', 'pop2'],
                  ['x7', 'clu2', 'pop3'],
                  ['x8', 'clu2', 'pop3']]
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='ICP')

        # missing CPI
        labels = [['A'], ['B'], ['C'], ['D']]
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='N')

        labels = [['A'], ['B'], ['C'], ['D']]
        struct = egglib.struct_from_iterable(labels, fmt='I')
        ctrl = {None: {None: {'A': [0], 'B': [1], 'C': [2], 'D': [3]}}}
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # repeated CPIN
        labels = [['A', 'A', 'A', 'A', 'A']]
        aln = egglib.Align(alphabet=egglib.Alphabet('char', ['Z'], []))
        aln.add_sample('A', 'Z')
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='NNIPC', data=aln)
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='NIIPC', data=aln)
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='NIPPC', data=aln)
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='NIPCC', data=aln)

        # (positive control)
        struct = egglib.struct_from_iterable(labels, fmt='*NIPC', data=aln)
        ctrl = {'A': {'A': {'A': [0]}}}
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # invalid number of items
        labels = [['x1', 'clu1', 'pop1'],
                  ['x2', 'clu1', 'pop1', 'x'],
                  ['x3', 'clu1', 'pop1'],
                  ['x4', 'clu1', 'pop1']]
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='ICP')

        labels = [['x1', 'clu1', 'pop1'],
                  ['x2', 'clu1', 'pop1'],
                  ['x3', 'clu1'],
                  ['x4', 'clu1', 'pop1']]
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='ICP')

        labels = [['A', 'A', 'A', '-'],
                  ['A', 'A', 'A', '-'],
                  ['A', 'A', 'A', '-'],
                  ['A', 'A', 'A', '-']]
        ctrl = {'A': {'A': {'A': [0, 1, 2, 3]}}}
        struct = egglib.struct_from_iterable(labels, fmt='ICP*')
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        labels = [['A', 'A', 'A', '-'],
                  'AAA-',
                  ['A', 'A', 'A', '-'],
                  ['A', 'A', 'A', '-']]
        with self.assertRaises(TypeError):
            error = egglib.struct_from_iterable(labels, fmt='ICP*')

        # comments
        labels = [['g', 'A', 'k', 'x'],
                  ['h', 'B', 'l', 'y'],
                  ['i', 'A', 'm', 'w'],
                  ['j', 'B', 'n', 'z']]
        ctrl = {None: {None: {'A': [0, 2], 'B': [1, 3]}}}
        struct = egglib.struct_from_iterable(labels, fmt='*I**')
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        ### data ###

        # name + data
        aln = egglib.Align(alphabet=egglib.Alphabet('int', [0,1], []))
        aln.add_sample('human', [0, 1, 1])
        aln.add_sample('cow', [0, 0, 1])
        aln.add_sample('pig', [0, 1, 1])
        aln.add_sample('dog', [1, 0, 0])
        aln.add_sample('cat', [1, 1, 0])

        labels = [['dog', 'house', 'eat bones'],
                  ['cat', 'house', 'eat mice'],
                  ['pig', 'farm', 'eat corn'],
                  ['human', 'house', 'eat burgers'],
                  ['cow', 'farm', 'eat grass']]
        ctrl = {None: {'house': {'human': [0], 'dog': [3], 'cat': [4]},
                       'farm': {'cow': [1], 'pig': [2]}}}
        struct = egglib.struct_from_iterable(labels, fmt='NP*', data=aln) # NB: names are taken as individual labels
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # name without data, data without name
        struct = egglib.struct_from_iterable(labels, fmt='NP*', data=aln) # control
        with self.assertRaises(ValueError):
            struct = egglib.struct_from_iterable(labels, fmt='NP*')
        struct = egglib.struct_from_iterable(labels, fmt='*P*') # control
        with self.assertRaises(ValueError):
            struct = egglib.struct_from_iterable(labels, fmt='*P*', data=aln)

        # name in labels but not in data
        labels = [['dog', 'house', 'eat bones'],
                  ['cat', 'house', 'eat mice'],
                  ['pig', 'farm', 'eat corn'],
                  ['velociraptor', 'jurassic park', 'eat goats'],
                  ['human', 'house', 'eat burgers'],
                  ['cow', 'farm', 'eat grass']]
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='NP*', data=aln)
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='NP*', data=aln, skip_missing_names=False)
        struct = egglib.struct_from_iterable(labels, fmt='NP*', data=aln, skip_missing_names=True)
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # name in data but not in labels
        labels = [['dog', 'house', 'eat bones'],
                  ['cat', 'house', 'eat mice'],
                  ['pig', 'farm', 'eat corn'],
                  ['velociraptor', 'jurassic park', 'eat goats'],
                  ['cow', 'farm', 'eat grass']]
        ctrl = {None: {'house': {'dog': [3], 'cat': [4]},
                       'farm': {'cow': [1], 'pig': [2]}}}
        struct = egglib.struct_from_iterable(labels, fmt='NP*', data=aln, skip_missing_names=True)
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        # duplicate in data
        aln.add_sample('cat', [1, 1, 1])
        del labels[3]
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='NP*', data=aln)

        labels.append(['cat', 'house', 'eat birds'])
        del aln[aln.ns-1]
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='NP*', data=aln)

        ### missing ###
        labels = [['a comment', 'demeA', 'mussel1'],
                  ['a comment', 'demeA', 'mussel2'],
                  ['a comment', 'demeB', 'mussel3'],
                  ['a comment', 'demeB', 'mussel4'],
                  ['a comment', 'demeA', 'mussel5'],
                  ['a comment', 'demeB', 'mussel6']]
        ctrl = {None: {'demeA': {'mussel1': [0], 'mussel2': [1], 'mussel5': [4]},
                       'demeB': {'mussel3': [2], 'mussel4': [3], 'mussel6': [5]}}}
        struct = egglib.struct_from_iterable(labels, fmt='*PI')
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        labels = [['a comment', 'demeA', 'mussel1'],
                  ['a comment', 'demeA', 'mussel2'],
                  ['a comment', 'demeB', 'mussel3'],
                  ['a comment', None, None],
                  ['a comment', 'demeA', 'mussel5'],
                  ['a comment', None, None]]
        ctrl = {None: {'demeA': {'mussel1': [0], 'mussel2': [1], 'mussel5': [4]},
                       'demeB': {'mussel3': [2]}}}
        struct = egglib.struct_from_iterable(labels, fmt='*PI')
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        aln = egglib.Align(alphabet=egglib.Alphabet('int', [0,1], []))
        aln.add_sample('mussel1', [0])
        aln.add_sample('mussel2', [0])
        aln.add_sample('mussel3', [0])
        aln.add_sample('mussel4', [1])
        aln.add_sample('mussel5', [1])
        aln.add_sample('mussel6', [1])
        aln.add_sample('mussel7', [1])
        struct = egglib.struct_from_iterable(labels, fmt='*PN', data=aln)
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})

        labels = [['a comment', 'demeA', 'mussel1'],
                  ['a comment', 'demeA', 'mussel2'],
                  ['a comment', 'demeB', 'mussel3'],
                  ['a comment', None, 'mussel4'],
                  ['a comment', 'demeA', 'mussel5'],
                  ['a comment', None, None]]
        struct = egglib.struct_from_iterable(labels, fmt='*PN', data=aln)
        ing, otg = struct.as_dict()
        self.assertDictEqual(ing, ctrl)
        self.assertDictEqual(otg, {})
        with self.assertRaises(ValueError):
            error = egglib.struct_from_iterable(labels, fmt='*PI')

        # str
        try:
            f, fname = tempfile.mkstemp()
            for row in labels:
                os.write(f, b'\t'.join([str(i).encode() for i in row]) + b'\n')
            os.close(f)
            with open(fname) as f:
                struct = egglib.struct_from_iterable(f, function=lambda x: x.strip().split('\t'), fmt='*PN', data=aln, missing='None')
            ing, otg = struct.as_dict()
            self.assertDictEqual(ing, ctrl)
            self.assertDictEqual(otg, {})
        finally:
            if os.path.isfile(fname): os.unlink(fname)

        ### start/stop ###
        # skip parts of list/file
        try:
            f, fname = tempfile.mkstemp()
            os.write(f, b'''\
#01 An example file
#02 This is a comment line
#03 Below header line
#04 Number,Individual name,pop,
#05,sample #01,pop 1,
#06,sample #02,pop 1,
#07,sample #03,pop 1,
#08,sample #04,pop 1,
#09,sample #05,pop 2,
#10,sample #06,pop 2,
#11,sample #07,pop 3,
#12,sample #08,pop 3,
#13,sample #09,pop 3,
#14,sample #10,pop 3,
''')
            os.close(f)
            ctrl = {None: {'pop 1': {'sample #01': [0], 'sample #02': [1],
                                     'sample #03': [2], 'sample #04': [3]},
                'pop 2': {'sample #05': [4], 'sample #06': [5]},
                'pop 3': {'sample #07': [6], 'sample #08': [7],
                          'sample #09': [8], 'sample #10': [9]}}}

            # start
            with open(fname) as f:
                struct = egglib.struct_from_iterable(f, function=lambda x: x.split(','), fmt='*IP*', start=4)
            ing, otg = struct.as_dict()
            self.assertDictEqual(ing, ctrl)
            self.assertDictEqual(otg, {})

            # out of bounds values supported
            with open(fname) as f:
                struct = egglib.struct_from_iterable(f, function=lambda x: x.split(','), fmt='*IP*', start=4, stop=28)
            ing, otg = struct.as_dict()
            self.assertDictEqual(ing, ctrl)
            self.assertDictEqual(otg, {})

            # stop
            ctrl = {None: {'pop 1': {'sample #02': [0], 'sample #03': [1], 'sample #04': [2]},
                'pop 2': {'sample #05': [3], 'sample #06': [4]},
                'pop 3': {'sample #07': [5], 'sample #08': [6]}}}
            with open(fname) as f:
                struct = egglib.struct_from_iterable(f, function=lambda x: x.split(','), fmt='*IP*', start=5, stop=12)
            ing, otg = struct.as_dict()
            self.assertDictEqual(ing, ctrl)
            self.assertDictEqual(otg, {})
        finally:
            if os.path.isfile(fname): os.unlink(fname)
